from agox.generators.symmetry_utils.symmetry_groups import (
    default_sym_dict,
    cluster_sym_dict,
)

import numpy as np
from ase.io import read, write
from ase.data import covalent_radii
from math import cos, sin, pi, tan, atan
import scipy.spatial.distance as ssd
from scipy.spatial.distance import cdist, pdist
from scipy.cluster.hierarchy import fcluster, linkage
from ase.geometry import wrap_positions, get_distances
from ase import Atoms, Atom
from ase.build.tools import sort


def upper_tri_masking(A):
    m = A.shape[0]
    r = np.arange(m)
    mask = r[:, None] < r
    return A[mask]


class SymmetryEnforcer:
    def __init__(
        self,
        slab,
        max_at=12,
        mode="cluster",
        group=None,
        confinement_cell=None,
        cell_corner=None,
        dimensionality=3,
    ):
        self.slab = slab
        self.group = group
        self.dimensionality = 3
        self.cell = self.slab.get_cell()
        self.pbc = slab.get_pbc()
        self.max_at = max_at
        # get cell vectors
        x = np.random.uniform(0, 1)
        self.radius = 1.5
        self.sym_type = None
        self.confinement_cell = confinement_cell
        self.cell_corner = cell_corner
        self.force_group = None
        self.cluster_center = None
        self.construction = np.random.choice(["outer", "inner"])
        self.mode = mode
        self.older_max_at = 0
        self.tries_max = 0
        # determine the geometry of the cell
        if not self.pbc.any():
            self.system = "cluster"
        elif (
            round(self.cell.angles()[2], 2) == 60
            or round(self.cell.angles()[2], 2) == 120
        ):
            self.system = "hexagonal"
        elif 88.5 < self.cell.angles()[2] < 91:
            if abs(self.cell.cellpar()[0] - self.cell.cellpar()[1]) < 1e-3:
                self.system = "square"
            else:
                self.system = "rectangular"
        elif abs(self.cell.cellpar()[0] - self.cell.cellpar()[1]) < 1e-3:
            self.system = "rhombic"
        else:
            self.system = "oblique"
        if self.pbc.all() and self.mode == "cluster":
            if self.system == "square":
                if abs(self.cell.cellpar()[0] - self.cell.cellpar()[2]) < 1e-3:
                    self.system = "cubic"
                else:
                    self.system = "tetragonal"
            elif self.system == "rectangular":
                self.system = "orthorombique"
        # axes :[[[[1,1,0],[1/2,1/2,0]]]] is an exemple of glide of 0.5x 0.5y after symmetry on axe 1,1,0
        # rotations:[[0,0,0],6] is a rotation aroud 0 0 1 on the center 0 0 0 of order 6
        # for cluster rotations defines de vector and the center is the center of the confinement cell
        # plane: defines a plane of symmetry and [-1,-1,-,1] is the inversion center
        self.dimensionality_angles = {
            3: {"theta": [0, 2 * np.pi], "phi": [0, np.pi]},
            2: {"theta": [0, 2 * np.pi], "phi": [np.pi / 2, np.pi / 2]},
            1: {"theta": [0, 0], "phi": [np.pi / 2, np.pi / 2]},
        }

        self.defaut_sym_dict = default_sym_dict
        self.cluster_sym_dict = cluster_sym_dict

    def get_box_vector(self):
        return self.confinement_cell.T @ np.random.rand(3) + self.cell_corner

    def get_displacement_vector(self, radius):
        theta = np.random.uniform(
            *self.dimensionality_angles[self.dimensionality]["theta"]
        )
        phi = np.random.uniform(*self.dimensionality_angles[self.dimensionality]["phi"])
        displacement = radius * np.array(
            [np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)]
        )
        return displacement

    def get_laplacian_matrix(self, relative_dist):
        A = np.logical_and(relative_dist > 0.75, relative_dist < 1.5).astype(int)
        D = np.diag(np.sum(A, axis=1))
        return D - A

    def get_cluster_center(self):
        center_of_conf = self.confinement_cell.T @ [0.5, 0.5, 0.5] + self.cell_corner
        if self.slab is None or len(self.slab) == 0:
            self.cluster_center = center_of_conf
            return center_of_conf
        else:
            at = self.return_three_closest(self.slab, center_of_conf)
            table = [
                self.slab[at[0]].position,
                1 / 2 * (self.slab[at[0]].position + self.slab[at[1]].position),
                1
                / 3
                * (
                    self.slab[at[0]].position
                    + self.slab[at[1]].position
                    + self.slab[at[2]].position
                ),
            ]
            self.cluster_center = np.random.choice(table)
            return self.cluster_center

    def get_mid_conf(self):
        len_clu = (
            self.confinement_cell.T @ [1, 1, 1] + self.cell_corner - self.cluster_center
        )
        return len_clu[0]

    def return_three_closest(self, candidate, position):
        # cov=0.7*(covalent_radii[self.atomic_number] + covalent_radii[self.atomic_number])
        p_new = np.copy(candidate.positions)
        dist = cdist(p_new, [position])
        dist = dist[dist > 0.05]
        idx = np.argpartition(dist, 3)
        return idx[:3]

    def check_dist_mic(self, candidate, mic=True):
        dMat = candidate.get_all_distances(mic=mic)
        numbers = [covalent_radii[candidate.numbers[i]] for i in range(len(candidate))]

        dist = upper_tri_masking(dMat)
        dist1 = np.tile(numbers, (len(numbers), 1))
        dist2 = dist1.T
        dist_tot_min = 0.7 * (dist1 + dist2)
        vect_dist_min = upper_tri_masking(dist_tot_min)

        new_vec = dist - vect_dist_min
        if any(new_vec < 0):
            return False
        return True

    def get_symmetry_points(
        self,
        p,
        combine_nearby_positions=False,
        sym_type="slab",
        order=None,
        atomic_number=None,
        candi=None,
    ):
        """The main function that returns on sym points altogether"""
        p = np.copy(p).reshape(1, 3)
        self.atomic_number = atomic_number
        self.sym_type = sym_type
        psym = np.zeros((1, 3))

        if sym_type == "cluster":
            if self.cluster_center is None:
                center = self.get_cluster_center()
            # axe0=p-center
            psym = p
            cov = 0.7 * (
                covalent_radii[self.atomic_number] + covalent_radii[self.atomic_number]
            )
            len_clu = self.get_mid_conf()
            psym = None
            tries = 0
            sym_dict = self.group
            prot = np.copy(p)
            p_new = np.copy(prot)
            # p_new=np.array([self.cluster_center + self.get_displacement_vector(self.radius)])
            self.construction = np.random.choice(["inner", "outer"])
            self.radius = cov / 0.7
            if self.construction == "outer":
                self.radius = len_clu
                p_new = np.array(
                    [self.cluster_center + self.get_displacement_vector(self.radius)]
                )
            radius0 = self.radius
            # radius0=len_clu/2
            # self.radius=radius0
            maxtries = 40
            for tries in range(maxtries):
                sym_dict = self.group
                if self.max_at < sym_dict["number_max"] or tries > maxtries / 2:
                    if tries == maxtries / 2:
                        # if self.max_at<=sym_dict['number_max']:
                        #    return None
                        # else:
                        self.radius = radius0
                    if sym_dict["points"] is not None:
                        p_new = self.select_mult_point(sym_dict, p_new)
                p_new = self.do_sym_operations(
                    p_new, sym_dict, cutoff=cov / 2, mic=False
                )
                if p_new.shape[0] > 2 and sym_dict["number_max"] > 2:
                    ghost = self.group["ghost"]
                    if self.pbc.all() and self.group["name"] in ["D2h", "D3h", "D4h"]:
                        ghost *= 2
                    if self.max_at < ghost or self.tries_max > 20:
                        group, tabl = self.select_group_desym_cluster(group=self.group)
                        if group["number_max"] > 6:
                            self.group = group
                            self.radius = radius0
                            continue
                        else:
                            return None
                    # dist = pdist(p_new)
                    dist = cdist(p_new, p_new)
                    dist = dist[dist > 0.01]
                    if candi is not None and p_new.shape[0] > 0:
                        vector, distances = get_distances(p_new, candi.positions)
                        dist = np.concatenate((dist.flatten(), distances.flatten()))
                    distances = dist
                    if (
                        2 < (self.max_at - p_new.shape[0]) < self.group["ghost"]
                        and tries < 10
                    ) or distances.shape[0] < 1:
                        continue
                    if (
                        np.amin(distances) < cov
                        or np.amin(distances) > 2.5 * cov
                        or p_new.shape[0] > self.max_at
                        or sym_dict["number_max"] % p_new.shape[0] != 0
                        or p_new.shape[0] < sym_dict["ghost"]
                    ):
                        if self.radius < len_clu:
                            if self.construction == "inner":
                                self.radius += (len_clu - cov) / (maxtries / 2)
                            else:
                                self.radius = max(
                                    cov,
                                    self.radius
                                    - (len_clu - 1.5 * cov) / (maxtries / 2),
                                )
                        p_new = np.array(
                            [
                                self.cluster_center
                                + self.get_displacement_vector(self.radius)
                            ]
                        )
                        continue
                elif (self.max_at > 2 and sym_dict["number_max"] > 2) or (
                    self.max_at == 2
                    and not self.group["name"].__contains__("C")
                    and p_new.shape[0] == 1
                    and self.system == "cluster"
                ):
                    p_new = np.array(
                        [
                            self.cluster_center
                            + self.get_displacement_vector(self.radius)
                        ]
                    )
                    continue
                p_new[:] = wrap_positions(p_new, self.cell, pbc=self.pbc)
                p_new = sort(p_new, tags=p_new[:, 2])
                p_new = sort(p_new, tags=abs(p_new[:, 2] - self.cluster_center[2]))
                if self.older_max_at == self.max_at:
                    self.tries_max += 1
                else:
                    self.tries_max == 0
                psym = np.copy(p_new)
                self.older_max_at = self.max_at
                break
        elif sym_type == "slab":
            cov = 0.75 * (
                covalent_radii[self.atomic_number] + covalent_radii[self.atomic_number]
            )
            psym = None
            tries = 0
            sym_dict = self.group
            prot = np.copy(p)
            p_new = np.copy(prot)
            while psym is None and tries < 15:
                if self.max_at < sym_dict["number_max"] or tries > 10:
                    if sym_dict["points"] is not None:
                        p_new = self.select_mult_point(sym_dict, p_new)
                p_new = self.do_sym_operations(
                    p_new, group=sym_dict, cutoff=0.25, mic=True
                )
                # advantage axis on rotation point because rotation point can come frome merging
                if (
                    sym_dict["axes"] is not None
                    and self.max_at >= sym_dict["number_max"] / 2
                ):
                    if p_new.shape[0] < sym_dict["number_max"] / 2:
                        tries += 1
                        continue
                p_new[:] = wrap_positions(p_new, self.cell, self.pbc)
                if p_new.shape[0] > 1:
                    vector, distances = get_distances(
                        p_new, p_new, cell=self.cell, pbc=self.pbc
                    )
                    distances = distances.flatten()
                    distances = distances[distances > 0.05]
                    if (
                        np.amin(distances) < cov
                        or p_new.shape[0] > self.max_at
                        or p_new.shape[0] > sym_dict["number_max"]
                    ):
                        p_new = np.array([self.get_box_vector()])
                        tries += 1
                        continue
                psym = np.copy(p_new)
        return psym

    def del_duplicate_positions(
        self, pos, cutoff=0.5, delete=True, merge=False, mic=False, atomic_number=1
    ):
        dists = pdist(pos, "euclidean")
        dup = np.nonzero(dists < cutoff)
        rem = np.array(self._row_col_from_pdist(pos.shape[0], dup[0]))
        # to test merge atoms
        tab_todel = None
        if delete:
            if rem.size != 0:
                try:
                    pos = np.delete(pos, rem[:, 0], axis=0)
                except:
                    pos = np.delete(pos, rem[0], axis=0)
            if merge:
                pos = self.get_symmetrised_positions(pos, atomic_number)
                return pos
            # slower method but more accurate used after to exec on a shorter list
            elif mic:
                # working with mic but slow then the shorter the list the better this method has been created for helping finding back the plane group
                tab_todel = None
                for i in range(len(pos - 1)):
                    disti = self.get_distances(pos, i, range(i + 1, len(pos)), mic=True)
                    tabdel = (np.argwhere(disti < cutoff) + i + 1).flatten()
                    if tabdel is not None and len(tabdel) > 0:
                        if tab_todel is None:
                            tab_todel = tabdel
                        else:
                            tab_todel = np.concatenate((tab_todel, tabdel))
                if tab_todel is not None and len(tab_todel) > 0:
                    tab_todel = np.unique(tab_todel)
                    pos = np.delete(pos, list(tab_todel), axis=0)
            return pos
        else:
            return rem

    def _row_col_from_pdist(self, dim, i):
        i = np.array(i)
        b = 1 - 2 * dim
        x = (np.floor((-b - np.sqrt(b**2 - 8 * i)) / 2)).astype(int)
        y = (i + x * (b + x + 2) / 2 + 1).astype(int)
        if i.shape:
            return list(zip(x, y))
        else:
            return [(x, y)]

    def rotate_around_vector(self, p_rot, theta, v, center=np.array([0, 0, 0])):
        if np.array(p_rot).ndim == 1:
            p_rot = [p_rot]
        norm = np.linalg.norm
        normv = norm(v)
        v /= normv
        p_copy = np.copy(p_rot)
        center = self.cell.cartesian_positions(center)
        p_copy = p_copy - center
        pos_new = None
        # for j, p1 in enumerate(p_rot):
        for a in theta:
            angle = a * np.pi / 180
            c = cos(angle)
            s = sin(angle)
            rot_new = (
                c * p_copy
                - np.cross(p_copy, s * v)
                + np.outer(np.dot(p_copy, v), (1.0 - c) * v)
                + center
            )
            if pos_new is None:
                pos_new = rot_new
            else:
                pos_new = np.vstack((pos_new, rot_new))
        return pos_new

    def get_distances(self, pos, a, i, mic=True, vector=False):
        p1 = pos[a]
        p2 = pos[i]

        cell = None
        pbc = None

        if mic:
            cell = self.cell
            pbc = self.pbc
        D, D_len = get_distances(
            np.array([p1]), np.array([p2]), cell=self.cell, pbc=self.pbc
        )

        if vector:
            D.shape = (-1, 3)
            return D
        else:
            D_len.shape = (-1,)
            return D_len

    def do_sym_operations(
        self,
        psym,
        group=None,
        cutoff=0.25,
        merge=False,
        mic=False,
        return_only_sym_pos=False,
    ):
        pcop = psym.copy()
        if self.group is not None and group is None:
            group = self.group
        else:
            group = group

        p_new = np.copy(psym)
        if p_new.ndim == 1:
            p_new = np.array([p_new])
        rot_list = group["rotations"]
        if "cluster" in group["system"]:
            mode = "cluster"
            centersym = self.cell.scaled_positions(self.cluster_center)
        else:
            mode = "slab"
        if group["name"] == "Ih5":
            for axe in group["axes"]:
                if mode == "cluster":
                    p_new = self.symmetry_ax(p_new, axe, p0=centersym)
                else:
                    p_new = self.symmetry_ax(p_new, axe)
        if rot_list is not None:
            for rotations in rot_list:
                order = rotations[1]
                listangles = [k * 360 // order for k in range(order)]
                centerot = rotations[0].copy()
                vecterot = [0, 0, 1]
                if mode == "cluster":
                    centerot = centersym
                    vecterot = rotations[0].copy()
                p_new = self.rotate_around_vector(
                    p_new, listangles, vecterot, center=centerot
                )
        p_new = self.del_duplicate_positions(
            p_new, cutoff=cutoff, delete=True, merge=merge, mic=mic
        )
        if self.pbc.any():
            p_new[:] = wrap_positions(p_new, self.cell, self.pbc)
        p_ref = None
        if group["axes"] is not None:
            for axe in group["axes"]:
                if mode == "cluster":
                    p_new = self.symmetry_ax(p_new, axe, p0=centersym)
                else:
                    p_new = self.symmetry_ax(p_new, axe)
        if mode == "cluster":
            if group["plane"] is not None:
                for plane in group["plane"]:
                    if group["plane"] is not None:
                        if plane == [-1, -1, -1]:
                            p_inv = -p_new + 2 * self.cell.cartesian_positions(
                                centersym
                            )
                            p_new = np.vstack((p_new, p_inv))
                        elif plane == [1, 1, 1]:
                            p_inv = p_new.copy()
                            p_inv = p_inv - (
                                2
                                * (
                                    p_inv * [0, 0, 1]
                                    - [
                                        0,
                                        0,
                                        self.cell.cartesian_positions(centersym)[2],
                                    ]
                                )
                            )
                            p_new = np.vstack((p_new, p_inv))
                        elif plane == [1, -1, -1]:
                            p_inv = p_new.copy()
                            p_inv = p_inv - (
                                2
                                * (
                                    p_inv * [0, 1, 1]
                                    - [
                                        0,
                                        self.cell.cartesian_positions(centersym)[1],
                                        self.cell.cartesian_positions(centersym)[2],
                                    ]
                                )
                            )
                            p_new = np.vstack((p_new, p_inv))
                        else:
                            p_new = self.symmetry_plane(p_new, plane, p0=centersym)
        if mode != "cluster":
            p_new[:] = wrap_positions(p_new, self.cell, self.pbc)
        p_sym2 = np.copy(p_new)
        if return_only_sym_pos:
            p_sym2 = self.del_duplicate_positions(
                p_sym2, cutoff=0.1, delete=True, merge=False, mic=True
            )
            jdel = []
            for j, sy in enumerate(p_sym2):
                if any((pcop[:] == sy).all(axis=0)):
                    jdel.append(j)
            psym_2 = np.delete(p_sym2, jdel, axis=0)
            return p_sym2
        p_sym2 = self.del_duplicate_positions(
            p_sym2, cutoff=cutoff, delete=True, merge=merge, mic=mic
        )
        return p_sym2

    def identify_mult_point_fast(
        self, candidate, table_mult, i, displacement, keep_sym=True
    ):
        # super fast identification that relies on direct coordinates
        ind = self.get_position_in_table(candidate, i, table_mult)
        if table_mult[ind] == self.group["number_max"]:
            if keep_sym == True or self.group["axes"] is None:
                return [0.0, 0.0, 0.0]
        # print(table_mult[ind])
        if (
            table_mult[ind] < int(self.group["number_max"] / 2) - 0.1
            or self.group["axes"] is None
            or self.group["name"] == "p2gg"
        ):
            return [0.0, 0.0, 1.0]
        else:
            pos = np.copy([candidate.positions[i]])
            pos[0, 2] = 0
            pos2 = np.copy(pos)
            ax = None
            if self.group["axes"] is not None:
                ax = self.is_on_ax2(pos)
                if ax is False:
                    if (
                        keep_sym == False
                        and table_mult[ind] == self.group["number_max"]
                    ):
                        return [0.0, 0.0, 0.0]
                    return ax
                ax = self.cell.cartesian_positions(ax)
                ax = ax / np.linalg.norm(ax)
                ax[2] = np.random.rand()
                return ax
        return displacement

    def identify_mult_point_cluster_new(
        self, candidate, table_mult, i, displacement, keep_sym
    ):
        ind = self.get_position_in_table(candidate, i, table_mult)
        if table_mult[ind] == self.group["number_max"]:
            # if keep_sym ==True:
            return [0.0, 0.0, 0.0]
        axes = self.group["axes"]
        plane = self.group["plane"]
        rotations = self.group["rotations"]
        pos = candidate[i].position.copy()
        mult = np.array(
            [
                1,
                1,
                1,
            ]
        )
        if candidate.pbc.all():
            for j, elem in enumerate(self.cell.scaled_positions(pos)):
                if elem <= 0.05 or 0.48 < elem < 0.52:
                    mult[j] = 0
        pos -= self.cluster_center
        pos *= mult
        if (
            np.linalg.norm(pos) < 1e-3
            or (2 < table_mult[ind] < self.group["ghost"])
            or table_mult[ind] == 1
        ):
            return False
        to_ret = np.random.choice((-1, 1)) * pos / np.linalg.norm(pos)
        if table_mult[ind] < self.group["number_max"] / 2 or axes is None:
            return to_ret
        elif table_mult[ind] == self.group["number_max"] / 2 and axes is not None:
            if self.group["name"].__contains__("h") is True and abs(pos[2]) < 1:
                to_ret = np.array(
                    [
                        np.random.random() * np.random.choice((-1, 1)),
                        np.random.random() * np.random.choice((-1, 1)),
                        0,
                    ]
                )
                to_ret /= np.linalg.norm(to_ret)
            else:
                to_ret = to_ret + np.random.choice(
                    (-1, 1)
                ) * np.random.random() * np.array([0, 0, 1])
                to_ret /= np.linalg.norm(to_ret)
            return to_ret
        return to_ret

    def identify_mult_point_cluster(
        self, candidate, table_mult, i, displacement, keep_sym
    ):
        ind = self.get_position_in_table(candidate, i, table_mult)
        if table_mult[ind] == self.group["number_max"]:
            # if keep_sym ==True:
            return [0.0, 0.0, 0.0]
        axes = self.group["axes"]
        plane = self.group["plane"]
        rotations = self.group["rotations"]
        pos = candidate[i].position.copy()
        pos -= self.cluster_center

        if np.linalg.norm(pos) < 1e-3 or 2 < table_mult[ind] < self.group["ghost"]:
            return False
        numofsym = 4
        if axes is not None and rotations is not None:
            numofsym = len(axes) * 2
        if self.group["name"] in ["D2d_2", "D2d", "D4d", "D6d"]:
            numofsym = 4
        if self.group["name"] == "D2":
            return pos / np.linalg.norm(pos)
        # handles simple C groups and D groups
        if axes is None and plane is None and rotations is not None:
            # C and D groups on z axis
            if table_mult[ind] <= 2:
                pos[0] = 0
                pos[1] = 0
            # D point groups on rot centers
            elif self.group["name"].__contains__("D") is True:
                # pos[0]=random.random()
                pos[2] = 0
            return pos / np.linalg.norm(pos)
        # handle is in the xy plane
        if table_mult[ind] <= self.group["number_max"] / numofsym or self.group[
            "name"
        ] in ["Th"]:
            if table_mult[ind] > 2 and self.group["name"].__contains__("D"):
                pos[2] = 0
            if abs(pos[2]) < 6e-1:
                pos[2] = 0
            return pos / np.linalg.norm(pos)
        # in the plane + axe  linear combination of vector - center and  (001) is in the symmetric plane
        axes = np.random.random() * pos + np.random.choice(
            (-1, 1)
        ) * np.random.random() * np.array([0, 0, 1])
        axes /= np.linalg.norm(axes)
        """
        if  table_mult[ind]>self.group['number_max']/numofsym and self.group['name'] not in ['Th','Td'] and pos[2]>3e-1 and axes is not None:
            axes=random.random()*pos+random.choice((-1, 1))*random.random()*np.array([0,0,1]) 
            axes/=np.linalg.norm(axes)
            #print(str(self.group['name'])+ ' bouger dans le plan')
        else:
            axes=pos
            axes/=np.linalg.norm(pos) 
        """
        return axes

    def switch_mult_point(self, pos, force_number=0):
        if np.array(pos).ndim == 1:
            pos = np.array([pos])
        mode = "slab"
        if self.group["name"].__contains__("p") == False:
            mode = "cluster"
        pos2 = np.copy(pos)
        sym_dict = self.group
        new_mult = self.select_mult_point(sym_dict, pos2)
        new_mult = self.do_sym_operations(new_mult, sym_dict)
        tries = 0
        if force_number > 0:
            while new_mult.shape[0] != force_number and tries < 50:
                new_mult = self.select_mult_point(sym_dict, pos2)
                new_mult = self.do_sym_operations(new_mult, sym_dict)
                tries += 1
                if mode == "cluster" or tries > 20:
                    pos2 = np.array([self.get_box_vector()])
            if tries == 49:
                return None
            return new_mult
        else:
            return new_mult

    def get_position_in_table(self, a, ind, table_mult):
        template = a.get_template()
        table = np.concatenate(([0], table_mult[:-1]))
        for i in range(len(table)):
            if i > 0:
                table[i] = table[i] + table[i - 1]
        table += len(template)
        table = table.tolist()
        for k in range(len(table) - 1):
            if table[k + 1] > ind >= table[k]:
                return k
        return len(table) - 1

    def get_two_mult_points(self, a, table_mult, atomic_number=None, n=2):
        template = a.get_template()
        table_mult2 = table_mult.copy()
        table = np.concatenate(([len(template)], table_mult2[:-1]))
        for i in range(1, len(table)):
            table[i] += table[i - 1]
        table = table.tolist()
        table_candi = []
        twopoints = []
        tries = 0
        copytabl = table_mult.copy()
        randax = np.random.random()

        while len(twopoints) < n and tries < 30:
            r = np.random.randint(0, len(table))
            tries += 1
            if atomic_number is None or atomic_number == a.numbers[int(table[r])]:
                if r not in twopoints:
                    if (
                        table_mult2[r] == self.group["number_max"]
                        and self.group["number_max"] > 3
                    ):
                        continue
                    if (
                        table_mult2[r] == self.group["number_max"] / 2
                        and self.group["axes"] is not None
                        and self.group["number_max"] > 4
                    ):
                        if n > 2 or randax < 0.7:
                            continue
                    twopoints.append(r)
                    if (
                        table_mult2[r] >= self.group["number_max"] / 2
                        and self.group["number_max"] > 2
                    ):
                        break
                    tries = 0
        twopoints.sort(reverse=True)
        for j in range(len(twopoints)):
            table_mult2.pop(twopoints[j])
        for k in twopoints:
            if k == len(table) - 1:
                table_candi += range(len(a) - copytabl[-1], len(a))
            else:
                table_candi += range(table[k], table[k + 1])
        return table_candi, table_mult2

    def fill_mult_point(self, pos, force_number=-1, candidate=None):
        # fill  a cell with special point
        sym_dict = self.group
        maxtries = 60
        if sym_dict["name"].__contains__("p") == False:
            mode = "cluster"
            maxtries = 80
            pos = np.array([self.get_box_vector()])
            if candidate is not None and candidate.pbc.any():
                mic = True
            else:
                mic = False
        else:
            mode = "slab"
            mic = True
        new_mult = np.array([], dtype=np.int64).reshape(0, 3)
        table_mult = []
        if candidate is None:
            new_mult = self.select_mult_point(sym_dict, pos)
            new_mult = self.do_sym_operations(new_mult, sym_dict, mic=mic)
            table_mult = [new_mult.shape[0]]
            dists = pdist(new_mult, "euclidean")
            if len(dists) > 0 and min(dists) < 1.8:
                new_mult = np.array([], dtype=np.int64).reshape(0, 3)
        tries = 0
        while new_mult.shape[0] != force_number and tries < maxtries:
            if tries == maxtries - 1:
                return None, None
            if mode == "cluster" or tries > 5:
                pos = np.array([self.get_box_vector()])
            app_mult = self.select_mult_point(sym_dict, pos)
            app_mult = self.do_sym_operations(app_mult, sym_dict, mic=mic)
            dists = pdist(app_mult, "euclidean")
            if len(dists) > 0 and min(dists) < 1.8:
                tries += 1
                continue
            if candidate is not None:
                _, distances = get_distances(
                    np.vstack((new_mult, candidate.positions)),
                    app_mult,
                    cell=self.cell,
                    pbc=candidate.pbc,
                )
            elif new_mult.shape[0] > 0:
                _, distances = get_distances(
                    app_mult, new_mult, cell=self.cell, pbc=[1, 1, 1]
                )
            if new_mult.shape[0] == 0:
                distances = np.array([2])
            distances = distances.flatten()
            if (
                new_mult.shape[0] + app_mult.shape[0] <= force_number
                and min(distances) > 1.8
            ):
                new_mult = np.vstack((new_mult, app_mult))
                table_mult.append(app_mult.shape[0])
            elif min(distances) > 1.8:
                new_mult = app_mult
                table_mult = [app_mult.shape[0]]
            tries += 1
        if new_mult.shape[0] != force_number:
            return None, None
        return new_mult, table_mult

    def select_mult_point(self, sym_dict, p_rot, number=1):
        """
        This function selects a random special point
        for cluster it places atoms according to the cluster center
        """
        if self.group is not None:
            group = self.group
        else:
            group = sym_dict
        if group["name"].__contains__("p") == False:
            mode = "cluster"
        else:
            mode = "slab"
        if np.array(p_rot).ndim == 1:
            p_rot = np.array([p_rot])
        if type(p_rot) == list:
            p_rot = np.array(p_rot)
        psym = sym_dict["points"]
        if self.pbc.all() and mode == "cluster":
            # add some wyckoff positions for bulk
            psym = np.vstack(
                (psym, self.cluster_center, self.cluster_center * [1, 0, 0])
            )
        x = np.random.uniform(0, 1)
        y = np.random.uniform(0, 1)
        if mode == "cluster":
            # else:
            p_rot = abs(p_rot) - self.cluster_center
            x = p_rot[0][0]
            y = p_rot[0][2]
        # if  self.sym_type=='cluster':
        select = np.random.randint(0, len(psym))
        psym = psym[select].copy()
        for j, elem in enumerate(psym):
            if type(elem) == int or type(elem) == float:
                continue
            if elem == "x":
                psym[j] = x
            elif elem == "y":
                psym[j] = y
            elif elem == "1/2(1-x)":
                psym[j] = 1 / 2 * (1 - x)
            elif elem.__contains__("x"):
                elem = elem.replace("x", str(x))
                elem = eval(elem)
                psym[j] = elem
        psym = np.copy(np.array(psym))
        psym = psym.astype(np.float64)
        if self.pbc.all() and mode == "cluster":
            for kl, ps in enumerate(psym):
                if ps == 0 and np.sum(psym) != 0:
                    psym[kl] = np.random.choice([0, self.cluster_center[kl]])
        if mode != "cluster":
            psym = np.array(self.cell.cartesian_positions(psym))
            if psym[2] == 0:
                try:
                    psym[:, 2] = p_rot[:, 2]
                    return psym
                except:
                    psym[2] = p_rot[0, 2]
        else:
            psym += self.cluster_center
        return np.array([psym])

    def is_on_ax2(self, pref, lim=5e-2):
        """
        This function identifies the mirror on which the atom is on
        based on sclaed positions
        """
        pos = self.cell.scaled_positions(pref)
        cell_u = self.cell.cellpar()[0]
        lim = lim * 12 / cell_u
        pos = pos[0]
        if self.group["name"] == "p4g":
            if 0.5 - lim <= abs(pos[1] - pos[0]) <= 0.5 + lim:
                return [1, 1, 0]
            if 0.5 - lim <= abs(pos[1] + pos[0] - 1) <= 0.5 + lim:
                return [1, -1, 0]
        if self.group["name"] == "p4m" or self.group["name"] == "p2mm":
            if 0.5 - lim <= abs(pos[1]) <= 0.5 + lim:
                return [1, 0, 0]
            if 0.5 - lim <= abs(pos[0]) <= 0.5 + lim:
                return [0, 1, 0]
        if self.group["name"] != "p3m1":
            if pos[0] < lim or pos[0] > 1 - lim:
                return [0, 1, 0]
            if pos[1] < lim or pos[1] > 1 - lim:
                return [1, 0, 0]
            if 1 - lim <= pos[0] + pos[1] <= 1 + lim:
                return [1, -1, 0]
        if self.group["name"] != "p31m":
            if pos[1] - lim <= pos[0] <= pos[1] + lim:
                return [1, 1, 0]
            if self.system == "hexagonal":
                if (
                    0.5 - 0.5 * pos[0] - lim <= pos[1] <= 0.5 - 0.5 * pos[0] + lim
                    or round(1 - 0.5 * pos[0], 2) - lim
                    <= pos[1]
                    <= round(1 - 0.5 * pos[0], 2) + lim
                ):
                    return [1, -1 / 2, 0]
                if (
                    round(0.5 - 0.5 * pos[1], 2) - lim
                    <= pos[0]
                    <= round(0.5 - 0.5 * pos[1], 2) + lim
                    or round(1 - 0.5 * pos[1], 2) - lim
                    <= pos[0]
                    <= round(1 - 0.5 * pos[1], 2) + lim
                ):
                    return [-1 / 2, 1, 0]
        return False

    def replace_on_ax(self, pref):
        """
        This function replaces atom on the closest mirror
        useful because relaxation can move them out
        """
        pos = self.cell.scaled_positions(pref)
        # pos=pos[0]
        minim = 1
        if self.group["name"] == "p4g":
            d = abs(0.5 - abs(pos[1] - pos[0]))
            if d < minim:
                minim = d
                minax = [1, 1, 0]
            d = abs(0.5 - abs(pos[1] - pos[0] - 1))
            if d < minim:
                minim = d
                minax = [1, -1, 0]
        if self.group["name"] == "p4m" or self.group["name"] == "p2mm":
            d = abs(0.5 - abs(pos[1]))
            if d < minim:
                minim = d
                minax = [1 / 2, 0, 0]
            d = abs(0.5 - abs(pos[0]))
            if d < minim:
                minim = d
                minax = [0, 1 / 2, 0]
        if self.group["name"] != "p3m1":
            d = abs(pos[0])
            if d < minim or abs(d - 1) < minim:
                minim = d
                minax = [0, 1, 0]
            d = abs(pos[1])
            if d < minim or abs(d - 1) < minim:
                minim = d
                minax = [1, 0, 0]
            d = abs(pos[1] + pos[0])
            if abs(d - 1) < minim:
                minim = d
                minax = [1, -1, 0]
        if self.group["name"] != "p31m":
            d = abs(pos[1] - pos[0])
            if d < minim:
                minim = d
                minax = [1, 1, 0]
            if self.system == "hexagonal":
                d = abs(pos[1] + 0.5 * pos[0] - 0.5)
                if d < minim:
                    minim = d
                    minax = [1, -1 / 2, 0]
                d = abs(pos[1] + 0.5 * pos[0] - 1)
                if d < minim:
                    minim = d
                    minax = [1, -1 / 2, 0]
                d = abs(pos[0] + 0.5 * pos[1] - 1)
                if d < minim:
                    minim = d
                    minax = [-1 / 2, 1, 0]
                d = abs(pos[0] + 0.5 * pos[1] - 0.5)
                if d < minim:
                    minim = d
                    minax = [-1 / 2, 1, 0]
        if self.group["name"] == "p4g":
            if pos[1] < 0.5:
                pos_new = np.array([pos[1], pos[1] + 0.5, pos[2]])
            else:
                pos_new = np.array([pos[1], pos[1] - 0.5, pos[2]])

        elif minax == [1, -1 / 2, 0] and pos[1] < 0.5:
            pos_new = np.array([pos[0] * minax[0], 0.5 + pos[0] * minax[1], pos[2]])
        elif minax == [-1 / 2, 1, 0] and pos[0] < 0.5:
            pos_new = np.array([pos[0] * minax[0] - 0.5, pos[0] * minax[1], pos[2]])
        elif 1 / 2 in minax and (
            self.group["name"] == "p4m" or self.group["name"] == "p2mm"
        ):
            if minax[0] == 1 / 2:
                pos_new = np.array([pos[0], 1 / 2, pos[2]])
            else:
                pos_new = np.array([1 / 2, pos[1], pos[2]])
        elif minax[0] != 0:
            pos_new = np.array([pos[0] * minax[0], pos[0] * minax[1], pos[2]])
        else:
            pos_new = np.array([pos[1] * minax[0], pos[1] * minax[1], pos[2]])
        pos_newcar = self.cell.cartesian_positions(pos_new)
        mincar = self.cell.cartesian_positions(minax)
        mincar = mincar / np.linalg.norm(mincar)
        return pos_newcar, mincar, pos_new, minax

    def choose_rotation_ax(self, group, order):
        tab_rot = None
        rotations = group["rotations"]
        for rot in rotations:
            if rot[1] == order:
                if tab_rot is None:
                    tab_rot = np.array([rot[0]])
                else:
                    tab_rot = np.vstack((tab_rot, rot[0]))

        if tab_rot is None:
            return None
        else:
            ind = list(range(tab_rot.shape[0]))
            chosen = np.random.choice(ind)
            return tab_rot[chosen]

    def replace_on_point_cluster(self, pos, j):
        """
        This function is used to replace atoms on a guessed symmetric position
        """
        group = self.group
        axes = self.group["axes"]
        if pos.ndim == 2:
            pos = pos[0] - self.cluster_center
        elif group["number_max"] > 6:
            return self.cluster_center
        """
        if j < group['ghost'] and group['name'] not in ['Th','Td','T','Oh4','Ih5']:
            if sum(pos)>0:
                pos=np.linalg.norm(pos)*np.array([0,0,1])
            else:
                pos=-np.linalg.norm(pos)*np.array([0,0,1])
        """
        if group["name"] in ["D2", "D3", "D4", "D5", "D6"]:
            if group["name"] == "D2" and j == 2:
                if max(abs(pos)) == abs(pos[2]):
                    pos[0] = 0
                    pos[1] = 0
                else:
                    if abs(pos[0]) < abs(pos[1]):
                        pos[0] = 0
                    else:
                        pos[1] = 0
                    pos[2] = 0
            elif j == 2:
                pos[0] = 0
                pos[1] = 0
            elif j == group["ghost"]:
                sym_di = self.cluster_sym_dict
                gr = str(group["name"]) + "h"
                for sym in sym_di:
                    if sym["name"] == gr:
                        pseudo_axes = np.array(sym["axes"])
                        break
                ind = list(range(len(pseudo_axes)))
                chosen = np.random.choice(ind)
                # posz=pos[2].copy()
                pos[2] = 0
                pos = np.linalg.norm(pos) * np.array(pseudo_axes[chosen])
                # pos[2]=posz
        elif group["name"] in ["D2h", "D2h_2", "D2d", "D2d_2"] and j == 2:
            if max(abs(pos)) == pos[2]:
                pos[0] = 0
                pos[1] = 0
            elif group["name"] in ["D2d", "D2d_2"]:
                pos = np.linalg.norm(pos) * np.array([0, 0, 1])
            else:
                pos[2] = 0
                ind = list(range(len(axes)))
                chosen = np.random.choice(ind)
                if sum(pos) > 0:
                    pos = np.linalg.norm(pos) * np.array(axes[chosen])
                else:
                    pos = -np.linalg.norm(pos) * np.array(axes[chosen])
        elif axes is not None and j >= group["ghost"]:
            # for  Dh et Dd point groups
            if (
                group["ghost"] == group["number_max"] / 4
                and j == group["ghost"]
                and group["name"] not in ["D2d", "D2d_2"]
            ):
                pos[2] = 0
            posz = pos[2]
            ind = list(range(len(axes)))
            chosen = np.random.choice(ind)
            if sum(pos) > 0:
                pos = np.linalg.norm(pos) * np.array(axes[chosen])
            else:
                pos = -np.linalg.norm(pos) * np.array(axes[chosen])
            pos[2] = posz
        elif group["rotations"] is not None:
            # doit gerer C et D
            rot = self.choose_rotation_ax(self.group, int(self.group["number_max"] / j))
            if rot is not None:
                if sum(pos) > 0:
                    pos = np.linalg.norm(pos) * np.array(rot)
                else:
                    pos = -np.linalg.norm(pos) * np.array(rot)
            else:
                return None
        elif j < group["ghost"] and group["name"] not in [
            "Th",
            "Td",
            "T",
            "Oh4",
            "Ih5",
        ]:
            if sum(pos) > 0:
                pos = np.linalg.norm(pos) * np.array([0, 0, 1])
            else:
                pos = -np.linalg.norm(pos) * np.array([0, 0, 1])
        else:
            print(group)
            print("point not found " + str(self.group["name"]))
        try:
            if abs(pos[2]) < 0.6:
                pos[2] = 0
            pos += self.cluster_center
            return pos
        except:
            print(str(group["name"]) + " " + str(pos))
            pass
        return None

    def symmetry_ax(self, p_new, axe, p0=[0, 0, 0]):
        glidaxe = [0, 0, 0]
        axe2 = np.array(axe)
        c = self.cell.reshape(3, 3)
        if axe2.ndim == 2 and axe2.shape[0] > 1:
            glidaxe = np.array(axe2[1])
            # p0=np.dot(p0,c)
            axe = np.array(axe2[0])
            if axe2.shape[0] > 2:
                p0 = axe2[2]
        p0 = self.cell.cartesian_positions(p0)
        p0[2] = p_new[0][2]
        pref = np.array([p_new[0]])
        norm = np.linalg.norm
        nvec = np.cross(axe, np.array([0, 0, 1]))
        nvec = np.dot(nvec / norm(nvec), c)
        glidaxe = self.cell.cartesian_positions(glidaxe)
        for j, pr in enumerate(p_new):
            proj = self.get_orthog_vec_from_plane(pr, nvec, p0)
            # print(str(axe) +' ' + str(nvec) +' ' + str(proj))
            p_reflected = pr - 2 * proj
            p_reflected += glidaxe
            pref = np.vstack((pref, p_reflected))
        if pref.shape[0] > 1:
            p_new = np.vstack((p_new, pref[1:]))
        # print(p_new)
        return p_new

    def symmetry_plane(self, p_new, axe, p0=[0.5, 0.5, 0.5]):
        norm = np.linalg.norm
        pref = np.array([p_new[0]])
        glidaxe = [0, 0, 0]
        c = self.cell.reshape(3, 3)
        p0 = self.cell.cartesian_positions(p0)
        nvec = axe
        if np.array(axe).ndim == 2:
            glidaxe = np.array(axe[1])
            # p0=axe[2]
            # p0=np.dot(p0,c)
            axe = axe[0]
        nvec = np.dot(nvec / norm(nvec), c)
        for j, pr in enumerate(p_new):
            p_reflected = self.mirror_point_in_plane(pr, nvec, p0=p0)
            pref = np.vstack((pref, p_reflected))
        if pref.shape[0] > 1:
            p_new = np.vstack((p_new, pref[1:]))
        return p_new

    def mirror_point_in_plane(self, p, nvec, p0=np.array([0, 0, 0])):
        proj = self.get_orthog_vec_from_plane(p, nvec, p0)
        p_reflected = p - 2 * proj
        return p_reflected

    def get_orthog_vec_from_plane(self, p, nvec, p0=np.array([0, 0, 0])):
        a = p - p0
        nvec_unit = (nvec / np.linalg.norm(nvec)).reshape(1, 3)
        proj_orthog = a @ nvec_unit.T @ nvec_unit
        return proj_orthog

    def make_clustering(self, pos, atomic_number):
        min_dist = 1.4 * covalent_radii[atomic_number]
        Natoms = pos.shape[0]
        a = Atoms(
            numbers=[atomic_number] * Natoms,
            positions=pos,
            pbc=self.pbc,
            cell=self.cell,
        )
        dMat = a.get_all_distances(mic=True)
        dVec = ssd.squareform(dMat)
        linked = linkage(dVec, "single", metric=None)
        clustering = fcluster(linked, t=min_dist, criterion="distance") - 1
        return clustering

    def select_group_desym_cluster(
        self,
        group=None,
        table_mult=None,
    ):
        # handle decreasing symmetry on slab
        sym_di = self.cluster_sym_dict
        sym_dict = []
        proba = []
        if group is None:
            group = self.group
        for elem in sym_di:
            if group is not None:
                if elem["name"] in group["subgroups"] and self.system in elem["system"]:
                    sym_dict.append(elem)
        if len(sym_dict) == 0:
            print("fail desym " + str(elem["name"]))
            return group, table_mult
        ind = list(range(len(sym_dict)))
        chosen = np.random.choice(ind)
        group = sym_dict[chosen]
        if table_mult is not None:
            table_mult = self.table_cluster_desym(group, table_mult)
        return group, table_mult

    def table_cluster_desym(self, group, table_mult):
        table_mult_cluster_desym = []
        for ta in table_mult:
            if ta > group["number_max"]:
                if ta == 12 and group["number_max"] == 8:
                    table_mult_cluster_desym.append(4)
                    table_mult_cluster_desym.append(8)
                elif ta == 12 and group["name"] in ["C5"]:
                    table_mult_cluster_desym.append(5)
                    table_mult_cluster_desym.append(5)
                    table_mult_cluster_desym.append(1)
                    table_mult_cluster_desym.append(1)
                elif ta == 12 and group["name"] in ["C5v"]:
                    table_mult_cluster_desym.append(10)
                    table_mult_cluster_desym.append(1)
                    table_mult_cluster_desym.append(1)
                else:
                    num_f = ta // group["number_max"]
                    if ta % group["number_max"] != 0:
                        table_mult_cluster_desym += [ta % group["number_max"]]
                    table_mult_cluster_desym += [group["number_max"]] * num_f
            elif group["number_max"] % ta != 0:
                if ta == 12 and group["name"] == "D5d":
                    table_mult_cluster_desym.append(2)
                    table_mult_cluster_desym.append(10)
                elif ta == 6 and group["number_max"] == 8:
                    table_mult_cluster_desym.append(2)
                    table_mult_cluster_desym.append(4)
                elif ta == 12 and group["number_max"] == 16:
                    table_mult_cluster_desym.append(8)
                    table_mult_cluster_desym.append(4)
                elif ta == 6 and group["number_max"] == 16:
                    table_mult_cluster_desym.append(4)
                    table_mult_cluster_desym.append(2)
                elif ta == 8 and group["number_max"] == 12:
                    table_mult_cluster_desym.append(4)
                    table_mult_cluster_desym.append(4)
                elif ta == 2 and group["name"] in [
                    "C3",
                    "C4",
                    "C5",
                    "C6",
                    "C3v",
                    "C4v",
                    "C5v",
                    "C6v",
                ]:
                    table_mult_cluster_desym.append(1)
                    table_mult_cluster_desym.append(1)
                else:
                    table_mult_cluster_desym.append(ta)
            elif ta == 2 and group["name"] in ["C4", "C6", "C4v", "C6v"]:
                table_mult_cluster_desym.append(1)
                table_mult_cluster_desym.append(1)
            elif ta == 8 and group["name"] == "Td":
                table_mult_cluster_desym.append(4)
                table_mult_cluster_desym.append(4)
            else:
                table_mult_cluster_desym.append(ta)
        return table_mult_cluster_desym

    def select_group_desym(self, group=None, table_mult=None):
        # handle decreasing symmetry on slab
        sym_di = self.defaut_sym_dict
        sym_dict = []
        proba = []
        for elem in sym_di:
            if group is not None:
                if self.system in elem["system"]:
                    if (
                        len(list(set(elem["name"]) & set(group["name"]))) > 1
                        or len(group["name"]) == 2
                        or group["name"] == "pcmm"
                    ):
                        if (
                            elem["number_max"] == group["number_max"] / 2
                            or elem["number_max"] == group["number_max"] / 3
                        ):
                            sym_dict.append(elem)
                            if elem["axes"] == None and elem["number_max"] > 4:
                                proba.append(2)
                            else:
                                proba.append(1)
        group_or = group.copy()
        if len(sym_dict) == 0:
            return group, table_mult
        proba = np.array(proba) / np.sum(proba)
        ind = list(range(len(sym_dict)))
        chosen = np.random.choice(ind, p=proba)
        group = sym_dict[chosen]
        table_mult_new = []

        for t in table_mult:
            if t > group["number_max"]:
                if t == group_or["number_max"]:
                    if group["number_max"] == group_or["number_max"] / 2:
                        table_mult_new.append(int(group["number_max"]))
                        table_mult_new.append(int(group["number_max"]))
                    elif group["number_max"] == group_or["number_max"] / 3:
                        table_mult_new.append(int(group["number_max"]))
                        table_mult_new.append(int(group["number_max"]))
                        table_mult_new.append(int(group["number_max"]))
                    else:
                        return group_or, table_mult
                else:
                    for i in range(t):
                        table_mult_new.append(1)
            else:
                table_mult_new.append(int(t))
        return group, table_mult_new

    def select_group(self, return_all=False, mode=None):
        "This function selects a start group"
        sym_di = self.defaut_sym_dict
        sym_dict = []
        number_max = 1
        if mode == "cluster":
            sym_di = self.cluster_sym_dict
            sym_dict = []
            number_max = 16
        for elem in sym_di:
            if self.system in elem["system"] or isinstance(self.force_group, str):
                if elem["name"] == self.force_group:
                    sym_dict.append(elem)
                    return sym_dict[0]
                elif isinstance(self.force_group, str):
                    continue
                elif (
                    mode == "cluster"
                    and elem["number_max"] >= number_max
                    and len(elem["name"]) >= 2
                ):
                    # get rid of some point groups at some conditions
                    if elem["name"] in ["D5d"]:
                        continue
                    if (self.max_at % 2 != 0 or self.system == "cubic") and elem[
                        "name"
                    ] in ["D4h", "D4d"]:
                        continue
                    if self.max_at % 5 != 0 and elem["name"] in ["D5h"]:
                        continue
                    # if self.max_at % 3 !=0 and elem['name'] in ['D6h','D6d']:
                    #   continue
                    if self.max_at % 7 != 0 and elem["name"] in [
                        "D7h",
                        "D7d",
                    ]:  # these are not very useful most of the time
                        continue
                    sym_dict.append(elem)
                elif mode != "cluster" and elem["number_max"] >= number_max:
                    sym_dict.append(elem)
                    if return_all is False:
                        number_max = elem["number_max"]
        if len(sym_dict) == 1:
            return sym_dict[0]
        chosen = np.random.randint(0, len(sym_dict))
        if return_all == False:
            sym_dict = sym_dict[chosen]
        return sym_dict

    def determine_plane_group(self, pos, cutoff=0.5):
        # the number of atoms is supposed to remain constant by operations
        sym = self.select_group(return_all=True)
        final_group = None
        for group in sym:
            pos_to_compare = self.do_sym_operations(
                pos, group, cutoff=cutoff, merge=True, mic=True
            )
            if pos_to_compare.shape[0] <= pos.shape[0]:
                if (
                    final_group is None
                    or final_group["number_max"] <= group["number_max"]
                ):
                    final_group = group
                    break
        return final_group

    def get_clusters(self, pos, atomic_number):
        clustering = self.make_clustering(pos, atomic_number)
        Nclusters = np.max(clustering) + 1
        clusters = [[] for i in range(Nclusters)]

        for i, ci in enumerate(clustering):
            clusters[ci].append(i)
        return clusters

    def get_symmetrised_positions(self, pos, atomic_number=None):
        clusters = self.get_clusters(pos, atomic_number)

        Natoms = pos.shape[0]
        a = Atoms(
            numbers=[atomic_number] * Natoms,
            positions=pos,
            pbc=self.pbc,
            cell=self.cell,
        )
        dVec_Mat = a.get_all_distances(mic=True, vector=True)

        pos_symm_all = []
        for cluster in clusters:
            p0 = pos[cluster[0]]
            pos_cluster = np.array(
                [p0 + dVec for dVec in dVec_Mat[cluster[0], cluster]]
            )
            pos_symm = pos_cluster.mean(axis=0)
            pos_symm_all.append(pos_symm)
        pos_symm_all = np.array(pos_symm_all)

        return pos_symm_all

    def swap_some_points(self, candidate, environement, table_mult):
        # redo this part to switch
        not_sym = candidate.get_meta_information("non_sym")
        if not_sym is None:
            not_sym = []
        tabelem = environement.get_numbers()
        output = []
        for x in tabelem:
            if x not in output and x not in not_sym:
                output.append(x)
        tabelem = output
        elem = np.random.choice(tabelem)
        counter = 0
        for el in table_mult:
            if el <= 4:
                counter += 1
        if self.group["name"].__contains__("p") == True:
            if counter > 2:
                n = np.random.choice([2, 3])
            else:
                n = np.random.choice([1, 2])
        elif min(table_mult) >= 8:
            n = np.random.choice([1, 2])
        elif self.group["name"].__contains__("6") == True:
            n = np.random.choice([2, 3, 6])
        elif self.group["name"].__contains__("5") == True:
            n = np.random.choice([2, 5])
        elif counter > 2:
            n = np.random.choice(list(range(2, counter + 2)))
        else:
            n = np.random.choice([1, 2])
        print(n)
        # n=2
        tab_tofil, table_mult2 = self.get_two_mult_points(
            candidate, table_mult, atomic_number=elem, n=n
        )
        candi = Atoms(
            numbers=candidate.numbers[:],
            positions=candidate.positions[:],
            cell=candidate.cell,
            pbc=candidate.pbc,
        )
        candi = candidate.copy()
        del candi[tab_tofil]
        if tab_tofil is not None and len(tab_tofil) >= 1:
            pos, tb = self.fill_mult_point(
                candidate.positions[np.random.choice(tab_tofil)],
                force_number=len(tab_tofil),
                candidate=candi,
            )
            # print(str(pos) + ' ' +str(tb))
            if pos is None or pos.shape[0] != len(tab_tofil):
                return None, None
            else:
                for sugg in pos:
                    # candidate.extend(Atom(elem, sugg))
                    candi.extend(Atoms(numbers=[elem], positions=[sugg]))
                table_mult2 += tb
        else:
            return None, None
        if not self.check_dist_mic(candi, mic=True):
            # write("POSCAR", candi, format="vasp")
            print("check_dist mic")
            return None, None
        return candi, table_mult2

    def resymmetrize(
        self,
        a,
        group,
        table_mult,
        cutoff=0.5,
        keep_sym=True,
        indices_to_rattle=[],
        not_sym=[],
    ):
        """
        This function resymmetrize the structure according to the right group
        It is supposed to handle gaining or losing symmetry
        """

        template = a.get_template()
        resym = None
        self.group = group
        table = np.concatenate(([0], table_mult[:-1]))
        if self.group["name"].__contains__("p") == False:
            print("cluster")
            mode = "cluster"
        else:
            mode = "slab"
        for i in range(len(table)):
            if i > 0:
                table[i] = table[i] + table[i - 1]
        table += len(template)
        new_table_mult = []
        for j, ta in enumerate(table):
            new_t = table_mult[j]
            if a[ta].number in not_sym:
                newpos = np.array([a.positions[ta]])
            # elif keep_sym==False and resym is not None and new_t==group['number_max'] and ta not in indices_to_rattle:
            elif (
                keep_sym == False and resym is not None and ta not in indices_to_rattle
            ):
                dist = [0]
                k = 0
                while min(dist) < cutoff and k < new_t:
                    newpos = self.do_sym_operations(
                        a.positions[ta + k], cutoff=cutoff, merge=False
                    )
                    dist = cdist(resym[-new_t:], newpos)
                    dist = dist.flatten()
                    k += 1
            else:
                newpos = self.do_sym_operations(
                    a.positions[ta], cutoff=cutoff, merge=False
                )
            if newpos.shape[0] > 1:
                newpos = self.del_duplicate_positions(
                    newpos,
                    cutoff=cutoff,
                    delete=True,
                    merge=True,
                    atomic_number=a[ta].number,
                )
            if new_t < group["number_max"]:
                # handle losing symmetry
                if newpos.shape[0] > table_mult[j]:
                    if mode == "cluster":
                        newpos = self.replace_on_point_cluster(newpos, table_mult[j])
                        if newpos is None:
                            print(
                                "fail replace "
                                + str(table_mult[j])
                                + " "
                                + str(group["name"])
                            )
                            return None, None
                        newpos = self.do_sym_operations(
                            newpos, cutoff=cutoff, merge=False
                        )
                    if newpos.shape[0] > table_mult[j]:
                        print(
                            "problem1 of losing sym "
                            + " a.ppos  "
                            + str(a.positions[ta])
                            + " apr switch "
                            + str(group["name"])
                            + " taille "
                            + str(newpos.shape[0])
                            + " "
                            + str(table_mult[j])
                            + " "
                            + str(keep_sym)
                        )
                        newpos, new_t = self.fill_mult_point(
                            a.positions[ta], force_number=table_mult[j]
                        )
                    if newpos is None:
                        print("check fill 1339")
                        return None, None
            if newpos.shape[0] < table_mult[j]:
                new_t = [newpos.shape[0]]
                # Handle gaining symmetry here
                if table_mult[j] % newpos.shape[0] != 0:
                    if self.cluster_center is None:
                        self.cluster_center = np.array([0, 0, 0])
                    # print('prob of gaining sym '+str(newpos[0]-self.cluster_center) + ' '+str(group['name'])+' taille '+str(newpos.shape[0]) + ' '+str(table_mult[j]) +' '+str(keep_sym))
                    # newpos=newpos[:int(table_mult[j]/2)]
                    new_mult, new_2 = self.fill_mult_point(
                        newpos, force_number=table_mult[j]
                    )
                    new_t = []
                    newpos = None
                else:
                    new_mult, new_2 = self.fill_mult_point(
                        newpos, force_number=table_mult[j] - newpos.shape[0]
                    )
                if new_mult is None:
                    new_mult, new_2 = self.fill_mult_point(
                        newpos, force_number=table_mult[j]
                    )
                    new_t = []
                    newpos = None
                if new_mult is None:
                    print("check full 1357")
                    return None, None
                if newpos is None:
                    newpos = new_mult
                else:
                    newpos = np.vstack((newpos, new_mult))
                for n in new_2:
                    new_t.append(n)
                if newpos.shape[0] != table_mult[j]:
                    return None, None
            if resym is None:
                resym = newpos
            elif newpos is not None:
                resym = np.vstack((resym, newpos))
            try:
                new_table_mult.extend(new_t)
            except:
                new_table_mult.append(new_t)

        # new_table_mult=[x for xs in new_table_mult for x in xs]
        return resym, new_table_mult

    def merger(self, a, index, group, table_mult, cutoff=0.5, disp=None):
        """
        This function handles completing the cell corresponding to the right table of positions
        """
        template = a.get_template()
        newpos = None
        resym = None
        self.group = group
        mode = None
        if self.group["name"].__contains__("p") == False:
            mode = "cluster"
        table = np.concatenate(([0], table_mult[:-1]))

        for i in range(len(table)):
            if i > 0:
                table[i] = table[i] + table[i - 1]
        table += len(template)
        new_table_mult = []
        for j, ta in enumerate(table):
            ta = int(ta)
            if ta == index:
                num_to_return = table_mult[j]
                newpos = self.do_sym_operations(
                    a.positions[ta], cutoff=cutoff, merge=False
                )
                if newpos.shape[0] > 1:
                    newpos = self.del_duplicate_positions(
                        newpos,
                        cutoff=cutoff,
                        delete=True,
                        merge=True,
                        atomic_number=a[ta].number,
                    )
                new_t = [newpos.shape[0]]
                if newpos.shape[0] > num_to_return:
                    if mode == "cluster":
                        newpos = self.replace_on_point_cluster(newpos, table_mult[j])
                        if newpos is None:
                            return None, None, None
                        newpos = self.do_sym_operations(
                            newpos, cutoff=cutoff, merge=False
                        )
                    else:
                        # print('check displacement ' + str(newpos[0]))
                        newpos, num = self.fill_mult_point(
                            newpos, force_number=num_to_return
                        )
                    if newpos is None:
                        return None, None, None
                    if newpos.shape[0] > num_to_return:
                        if self.cluster_center is None:
                            self.cluster_center = np.array([0, 0, 0])
                        # print('problem in merge' + str(group['name'])+ ' '+str(table_mult[j]) + ' '+str(newpos.shape[0])+' '+str(newpos[0]-self.cluster_center)+' '+str(disp))
                        newpos = newpos[: table_mult[j]]
                    break
                elif newpos.shape[0] < table_mult[j]:
                    new_mult, new_2 = self.fill_mult_point(
                        newpos, force_number=table_mult[j] - newpos.shape[0]
                    )
                    if new_mult is None or newpos is None:
                        new_mult, new_2 = self.fill_mult_point(
                            newpos, force_number=table_mult[j]
                        )
                        new_t = []
                        if new_mult is None:
                            return None, None, None
                        else:
                            newpos = new_mult
                    else:
                        newpos = np.vstack((newpos, new_mult))
                    for n in new_2:
                        new_t.append(n)
                    new_table_mult = table_mult[:j] + new_t
                    if len(table_mult[j + 1 :]) > 0:
                        new_table_mult = new_table_mult + table_mult[j + 1 :]
                    if sum(new_table_mult) != sum(table_mult):
                        return None, None, None
                    # return newpos[:num_to_return] , new_table_mult , new_t
                break
        if newpos is None:
            return None, None, None
        return newpos[:num_to_return], table_mult, new_t

    """
    def set_single_position(self, a, index, pos_new, index_avail):
        indices_symm = set(self.get_symmetry_equivalent_atoms(a, index))
        indices_rattle = list(indices_symm.intersection(set(index_avail)))
        #print('indices_rattle:', indices_symm, indices_rattle)

        pos_symm = self.get_symmetry_points(pos_new)
        Nsym = pos_symm.shape[0]
        if Nsym > len(indices_rattle):
            return None
        indices_use = indices_rattle[:Nsym]
        a.positions[indices_use] = pos_symm
        return indices_use
    """



import pytest
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.descriptors import Fingerprint
import numpy as np
from agox.test.test_utils import test_data_dicts, get_test_environment, get_test_data

@pytest.mark.ray
@pytest.mark.parametrize('test_data_dict', test_data_dicts)
@pytest.mark.parametrize('train', [True, False])
def test_ray_global_gpr(test_data_dict, train, ray_fix, cmd_options):

    # Tolerances:
    tolerance = cmd_options['tolerance']

    path = test_data_dict['path']
    remove = test_data_dict['remove']

    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    train_data = data[0:25]
    test_data = data[25:35]

    # Make a global model: 
    descriptor = Fingerprint(environment=environment)

    # If the kernel hyperparameters are allowed to take extreme values (large amplitude, small length scale), 
    # it can ocassionally fail - even when the hyperparameters are properly transferred to the actors.
    # But I think that has to be a numerical issue? 
    # kernel = C(1, (1, 100)) * RBF(10, (10, 50)) + Noise(0.01)
    kernel = C() * RBF() + Noise(0.01)

    model = SparseGPR(descriptor=descriptor,
                kernel=kernel,
                n_optimize=2,
                use_ray=True,
                unc_jitter=1e-3,
                centralize=True,
                train_uncertainty=True)

    model.train(train_data)
    model.pool.synchronize_module(model)

    # Check properties on actors: 
    if train:
        result_dicts = model.pool.get_module_attributes(model, model.dynamic_attributes)

        for result_dict in result_dicts:
            for attribute_name in ['X', 'Xm', 'Kmm_inv', 'alpha', 'C_inv']:
                np.testing.assert_allclose(result_dict[attribute_name], model.__getattribute__(attribute_name), **tolerance)

            # Kernel check seperately because it is annoying. 
            np.testing.assert_allclose(result_dict['kernel'].theta, model.kernel.theta, **tolerance)

    # Predictions without Ray: 
    energies = np.array(model.predict_energy(test_data))
    forces = np.array(model.predict_forces(test_data))
    uncertainties = np.array(model.predict_uncertainty(test_data))
    unceertainty_forces = np.array(model.predict_uncertainty_forces(test_data))

    # Predictions with Ray:
    def predict(model, atoms): 
        E = np.array(model.predict_energy(atoms))
        F = np.array(model.predict_forces(atoms))
        sigma = np.array(model.predict_uncertainty(atoms))
        sigma_F = np.array(model.predict_uncertainty_forces(atoms))

        results = {'E': E, 'F': F, 'sigma': sigma, 'sigma_F': sigma_F}
        return results
    pool = model.pool
    # Make predictions on all actors:
    key = pool.get_key(model)
    results = pool.execute_on_actors(predict, [key], [test_data], {})

    # Check that all predictions on every actor are the same:
    results_actor_0 = results[0]
    for key in results[0].keys():
        for results_actor in results[1:]:
            np.testing.assert_allclose(results_actor_0[key], results_actor[key])

    # # Check that predictions are the same as without Ray:
    np.testing.assert_allclose(energies, results_actor_0['E'], **tolerance)
    np.testing.assert_allclose(forces, results_actor_0['F'], **tolerance)
    np.testing.assert_allclose(uncertainties, results_actor_0['sigma'], **tolerance)
    np.testing.assert_allclose(unceertainty_forces, results_actor_0['sigma_F'], **tolerance)



import os
from time import sleep
import numpy as np
from ase.io import write, read
from agox.writer import agox_writer
from agox.samplers.metropolis import MetropolisSampler

from ase.calculators.singlepoint import SinglePointCalculator

from agox.observer import Observer 

def safe_write(filename, atoms):
    write(filename=filename+'.lock', images=atoms, format='traj')
    os.rename(filename+'.lock', filename)

class ConcurrentTemperingSampler(MetropolisSampler):
    """
    Parallel tempering sampler for concurrent searches with multiple worker jobs. Written to work with ConcurrentDatabase, no guarantee that it works in any other setting. 

    Parameters
    ----------
    temperatures : list
        List of temperatures for each worker.
    database : ConcurrentDatabase
        The database that the sampler will attach to.
    swap_frequency : int
        The frequency at which the workers will swap candidates.
    gets : list
        List of dictionaries with the gets that the sampler will observe.
    sets : list
        List of dictionaries with the sets that the sampler will observe.
    swap_order : int
        Order of the observer.
    """
    
    name = 'ParallelTemperingSampler'

    def __init__(self, temperatures=[], database=None, swap_frequency=10, gets=[{'get_key':'evaluated_candidates'}, {}], 
                sets=[{}, {}], swap_order=2, **kwargs):
        super().__init__(gets=gets, sets=sets, **kwargs)
        self.swap_frequency = swap_frequency
        self.temperatures = temperatures
        self.temperature = temperatures[database.worker_number]
        self.swap_order = swap_order
        self.swap_func = self.metropolis_hastings_swap
        self.most_recently_accepted = {worker_number:None for worker_number in range(database.total_workers)}
        self.most_recently_accepted_iteration = {worker_number:0 for worker_number in range(database.total_workers)}

        self.add_observer_method(self.swap_candidates, sets=self.sets[1], gets=self.gets[1], order=self.swap_order, handler_identifier='database')

        self.attach_to_database(database)

    @agox_writer
    @Observer.observer_method
    def swap_candidates(self, database, state):
        """
        Assumes that the database is synced. 
        """
        
        if self.decide_to_swap(database):
            for candidate in database.candidates:

                worker_number = candidate.get_meta_information('worker_number')
                iteration = candidate.get_meta_information('iteration')
                accepted = candidate.get_meta_information('accepted')
                
                if accepted:
                    if iteration > self.most_recently_accepted_iteration[worker_number]:
                        self.most_recently_accepted[worker_number] = candidate
                        self.most_recently_accepted_iteration[worker_number] = candidate.get_meta_information('iteration')

            if self.verbose:
                energies = [self.most_recently_accepted[c].get_potential_energy() \
                            if self.most_recently_accepted[c] is not None else 0 for c in self.most_recently_accepted]
                if database.worker_number == 0:
                    text = 'PARALLEL TEMPERING:'  + f'{self.temperature:8.3f}, ' + ','.join([f'{e:8.3f}' for e in energies])
                else:
                    text = 'PARALLEL TEMPERING (not main worker):'
                self.writer(text)

            self.swap_func(database)
        
    def metropolis_hastings_swap(self, database):
        worker_number = database.worker_number
        total_workers = database.total_workers
        iteration = self.get_iteration_counter()

        # I abuse the filenames a bit here: 
        filename = database.filename[:-3] + '_swap_iteration_{}_worker_{}.traj'

        if worker_number == 0:
            # This one does the calculation and 'broadcasts' to the others over disk. 

            # Starting from the bottom: 
            for i in range(total_workers-1):                
                C_i = self.most_recently_accepted[i]
                C_j = self.most_recently_accepted[i+1]

                E_i = C_i.get_potential_energy()
                E_j = C_j.get_potential_energy()

                beta_i = 1/self.temperatures[i]
                beta_j = 1/self.temperatures[i+1]

                P = np.min([1, np.exp((beta_i-beta_j)*(E_i-E_j))])

                r = np.random.rand()

                if r < P:
                    self.writer('Swapped {} with {}'.format(i, i+1))
                    self.most_recently_accepted[i], self.most_recently_accepted[i+1] = self.most_recently_accepted[i+1], self.most_recently_accepted[i]
                    self.most_recently_accepted_iteration[i], self.most_recently_accepted_iteration[i+1] = iteration, iteration
                    
                else:
                    self.writer('Did not swap {} for {}'.format(i, i+1))

            # Write the candidates:
            for wn in range(1, total_workers):
                atoms = self.most_recently_accepted[wn]
                atoms.info = atoms.meta_information
                safe_write(filename.format(iteration, wn), atoms)
            self.chosen_candidate = self.most_recently_accepted[worker_number]

        else: 
            while not os.path.exists(filename.format(iteration, worker_number)):
                sleep(1)
            
            chosen_atoms = read(filename.format(iteration, worker_number))
            
            self.sample[0] = self.convert_to_candidate_object(chosen_atoms, self.sample[0].template)
            self.sample[0].meta_information = chosen_atoms.info
            scp = SinglePointCalculator(self.sample[0], energy=chosen_atoms.get_potential_energy(), forces=chosen_atoms.get_forces())
            self.sample[0].set_calculator(scp)

            os.remove(filename.format(iteration, worker_number))

    def convert_to_candidate_object(self, atoms_type_object, template):
        candidate =  self.candidate_instanstiator(template=template, positions=atoms_type_object.positions, numbers=atoms_type_object.numbers, 
                                          cell=atoms_type_object.cell, info=atoms_type_object.info)
        return candidate

    def decide_to_swap(self, database):
        return (self.get_iteration_counter() % self.swap_frequency == 0) * (database.total_workers > 1)    



from typing import List

import numpy as np
from ase import Atoms

from agox.utils.filters.ABC_filter import FilterBaseClass
from agox.models.descriptors.voronoi import Voronoi

class VoronoiFilter(FilterBaseClass):
    """
    Filther that removes structures if their graph has produced
    a given number of children.

    """

    name = "VoronoiGraphFilter"

    def __init__(self, max_number_of_children = 5, descriptor = None, **kwargs):
        """
        Parameters
        ----------
        max_number_of_children: int
            Maximum number of children.
        voronoi: VoronoiSampler
            VoronoiSampler object to calculate graphs.
        remove_tree: bool
            If True, the tree will be removed.
        """
                
        super().__init__(**kwargs)
        self.max_number_of_children = max_number_of_children

        if descriptor is None:
            self.descriptor = Voronoi(covalent_bond_scale_factor = 1.3,
                n_points = 8, angle_from_central_atom = 20)
        else:
            self.descriptor = descriptor
        
        #Tree stuff
        self.number_of_structures_in_last_iteration = 0
        self.number_of_children = {}
        self.graph_to_uuid = {}
        self.uuid_to_graph = {}

        self.uuids_to_remove = []
        self.removed_graphs = []        

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        """
        Filter the atoms object.

        Parameters
        ----------
        atoms
            The atoms object to be filtered.

        Returns
        -------
        indexes: array-like
            The indexes of the atoms that are kept.
        """

        if self.descriptor.template is None:
            self.descriptor.template = atoms[0].template

        graphs_to_remove = []
        new_uuids = []

        for i, structure in enumerate(atoms[self.number_of_structures_in_last_iteration:]):
            my_uuid = structure.get_meta_information('uuid')
            new_uuids.append(my_uuid)

            if structure.get_meta_information('final') is not True:
                continue

            my_graph = self.descriptor.create_features(structure, False)

            for uuid in new_uuids:
                self.uuid_to_graph[uuid] = my_graph

            self.graph_to_uuid[my_graph] = self.graph_to_uuid.get(my_graph, []) + new_uuids

            if my_graph in self.removed_graphs:
                self.uuids_to_remove += new_uuids

                if structure.has_meta_information('parent_uuids'):
                    parent_uuids = structure.get_meta_information('parent_uuids')
                    for parent_uuid in parent_uuids:
                        parent_graph = self.uuid_to_graph[parent_uuid]
                        graphs_to_remove.append(parent_graph)
                        if parent_graph not in self.removed_graphs:
                            self.removed_graphs.append(parent_graph)
                        self.uuids_to_remove.append(parent_uuid)

            if structure.has_meta_information('parent_uuids'):

                parent_uuids = structure.get_meta_information('parent_uuids')
                parent_search_iterations = structure.get_meta_information('parent_search_iterations')
                for parent_uuid, parent_search_iteration in zip(parent_uuids, parent_search_iterations):
                    parent_graph = self.uuid_to_graph[parent_uuid]

                    self.number_of_children[parent_graph] = self.number_of_children.get(parent_graph, 0) + 1
            
                    if self.number_of_children[parent_graph] >= self.max_number_of_children:
                        graphs_to_remove += [parent_graph]

            new_uuids = []

        for g in graphs_to_remove:
            if g not in self.removed_graphs:
                self.removed_graphs.append(g)
            self.uuids_to_remove += self.graph_to_uuid[g]

        self.uuids_to_remove = list(set(self.uuids_to_remove))

        indexes = [i for i in range(len(atoms)) if atoms[i].get_meta_information('uuid') not in self.uuids_to_remove]

        self.number_of_structures_in_last_iteration = len(atoms)
        return np.array(indexes)


from typing import List

import numpy as np
from ase import Atoms

from agox.utils.filters.ABC_filter import FilterBaseClass

class KMeansEnergyFilter(FilterBaseClass):
    name = "KMeansEnergyFilter"

    """
    Filter that removes all structures with energy above a certain threshold.
    
    Parameters
    ----------
    sample_size: int
        Number of structures to sample.
    max_energy: float
        Maximum energy above minimum energy.
    """

    def __init__(self, sample_size: int = 10, max_energy: float = 5, **kwargs):

        super().__init__(**kwargs)
        self.sample_size = sample_size
        self.max_energy = max_energy

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        """
        Filter out structures with energies higher than e_min + max_energy

        Parameters
        ----------
        all_structures : list of ase.Atoms or agox.candidates.candidate.Candidate

        Returns
        -------
        indices: np.ndarray
            Array of indices for structures to keep.
        """
        # Get energies of all structures:
        e_all = np.array([s.get_potential_energy() for s in atoms])
        e_min = min(e_all)

        for i in range(5):
            filt = e_all <= e_min + self.max_energy * 2**i
            if np.sum(filt) >= 2 * self.sample_size:
                break
        else:
            filt = np.ones(len(e_all), dtype=bool)
            index_sort = np.argsort(e_all)
            filt[index_sort[2 * self.sample_size :]] = False

        # The final set of structures to consider:
        indices = np.array([i for i in range(len(atoms)) if filt[i]])

        return indices


import pytest
import numpy as np
from agox.models.descriptors import SOAP
from agox.test.test_utils import test_data_dicts, get_test_environment, get_test_data, save_expected_data, load_expected_data, test_folder_path

@pytest.mark.parametrize('test_data_dict', test_data_dicts)
def test_descriptor(test_data_dict, cmd_options):

    create_mode = cmd_options['create_mode']
    tolerance = cmd_options['tolerance']

    path = test_data_dict['path']
    remove = test_data_dict['remove']
    name = test_data_dict['name']
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    species = set(data[0].get_chemical_symbols())
    descriptor = SOAP.from_species(species)

    F = np.array(descriptor.get_features(data))

    name = f'{test_folder_path}/descriptor_tests/expected_outputs/{descriptor.name}_{name}.pckl'
    
    if create_mode:
        save_expected_data(name, F)
    else:
        F_exp = load_expected_data(name)
        np.testing.assert_allclose(F, F_exp, **tolerance)







import numpy as np
import os
from agox.databases import Database
from argparse import ArgumentParser
from ase.io import read, write

def add_convert_parser(subparsers):
    parser = subparsers.add_parser('convert', help='Convert database files to trajectories.')
    parser.set_defaults(func=cli_convert)
    parser.add_argument('files', type=str, nargs='+')
    parser.add_argument('-n', '--name', type=str, default=None)

def convert_database_to_traj(database_path):
    """
    Convert database to list of ASE Atoms objects. 

    Currently assumes that the database can be read using the standard database 
    class (agox/modules/databases/database.py).

    Parameters
    -----------
    database_path: str
        Path to database on disk. 

    Returns
    --------
    list
        List of ASE Atoms objects. 
    """

    if os.path.exists(database_path):
        database = Database(database_path, initialize=False)
    else:
        print(f'Trajectory not found: {database_path}')
        return []

    trajectory = database.restore_to_trajectory()

    return trajectory

def find_file_types(paths):
    """
    Find file extension of given paths. 

    Parameters
    -----------
    paths: list of str
        Paths to check
    
    Returns 
    --------
    list: 
        List of strs indicating of the file types (extensions) of the given paths. 
    """
    file_types = []
    for path in paths:
        file_types.append(path.split('.')[-1])
    return file_types

def cli_convert(args):
    # Determine the file extension & ensure all paths have the same extension. 
    file_types = find_file_types(args.files)
    assert np.array([file_type == file_types[0] for file_type in file_types]).all()
    file_type = file_types[0]

    # If database files - convert to trajectory. 
    trajectory = []
    if file_type == 'db':
        for path in args.files:
            print(f'Converting: {path}')
            trajectory += convert_database_to_traj(path)

        # Naming: 
        name = args.name
        if name is None:
            name = 'converted_db.traj'
        
        # Save: 
        print(f'Saving trajectory file {name}')
        write(name, trajectory)





from .ABC_evaluator import EvaluatorBaseClass
import numpy as np
from agox.observer import Observer
from ase.calculators.singlepoint import SinglePointCalculator
from agox.writer import agox_writer

from ase.optimize.bfgs import BFGS
from ase.constraints import FixAtoms

from uuid import uuid4
import traceback

class LocalOptimizationEvaluator(EvaluatorBaseClass):

    """
    Evaluator that performs a local optimization of a candidate using a ASE BFGS optimizer.

    Parameters:
    -----------
    calculator: ASE calculator
        The calculator to use for the evaluation.
    optimizer: ASE optimizer, optional
        The optimizer to use for the local optimization. Default is BFGS.
    optimizer_run_kwargs: dict, optional
        The keyword arguments to pass to the optimizer.run() method.
    optimizer_kwargs: dict, optional
        The keyword arguments to pass to the optimizer constructor.
    fix_template: bool, optional
        Whether to fix the template atoms during the optimization. Default is True.
    constraints: list, optional
        List of constraints to apply during the optimization.
    store_trajectory: bool, optional
        Whether to store the trajectory of the optimization. Default is True.
        If False only the last step is stored.
    """

    name = 'LocalOptimizationEvaluator'

    def __init__(self, calculator, optimizer=BFGS, optimizer_run_kwargs={'fmax':0.25, 'steps':200},
                 optimizer_kwargs={'logfile':None}, fix_template=True, constraints=[],
                 store_trajectory=True, **kwargs): 
        super().__init__(**kwargs)
        self.calculator = calculator
        self.store_trajectory = store_trajectory
        
        # Optimizer stuff:
        self.optimizer = optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimizer_run_kwargs = optimizer_run_kwargs

        # Constraints:
        self.constraints = constraints
        self.fix_template = fix_template

        if self.name == 'LocalOptimizationEvaluator':
            if optimizer_run_kwargs['steps'] == 0:
                print(f'{self.name}: Running with steps = 0 is equivalent to SinglePointEvaluator, consider using that instead to simplify your runscript.')
                self.store_trajectory = False

    def evaluate_candidate(self, candidate):
        candidate.calc = self.calculator

        try:
            if self.optimizer_run_kwargs['steps'] > 0:
                self.apply_constraints(candidate)
                optimizer = self.optimizer(candidate, **self.optimizer_kwargs)

                # The observer here stores all the steps if 'store_trajectory' is True.
                optimizer.attach(self._observer, interval=1, candidate=candidate, steps=optimizer.get_number_of_steps)
        
                optimizer.run(**self.optimizer_run_kwargs)                
                candidate.add_meta_information('relax_index', optimizer.get_number_of_steps())

                # If 'store_trajectory' is False we manually store the last step.
                if not self.store_trajectory:
                    self.evaluated_candidates.append(candidate)
            else:
                F = candidate.get_forces(apply_constraint=False)
                E = candidate.get_potential_energy(apply_constraint=False)
                self.evaluated_candidates.append(candidate)
            
        except Exception as e:
            self.writer('Energy calculation failed with exception: {}'.format(e))
            traceback.print_exc()
            return False

        E = candidate.get_potential_energy(apply_constraint=False)
        F = candidate.get_forces(apply_constraint=False)
        self.writer(f'Final energy of candidate = {E:5.3f}')
        calc = SinglePointCalculator(candidate, energy=E, forces=F)
        candidate.calc = calc

        return True


    def _observer(self, candidate, steps):
        E = candidate.get_potential_energy(apply_constraint=False)
        F = candidate.get_forces(apply_constraint=False)
        
        traj_candidate = candidate.copy()
        calc = SinglePointCalculator(traj_candidate, energy=E, forces=F)
        traj_candidate.set_calculator(calc)
        traj_candidate.add_meta_information('relax_index', steps())
        traj_candidate.add_meta_information('final', False)
        if steps() > 0:
            traj_candidate.add_meta_information('uuid', str(uuid4()))
        self.writer(f'Step {steps()}: {E:.3f}')

        if self.store_trajectory:
            self.evaluated_candidates.append(traj_candidate)

        #self.add_to_cache(self.set_key, [traj_candidate], mode='a')
    

    def apply_constraints(self, candidate):
        constraints = [] + self.constraints
        if self.fix_template:
            constraints.append(self.get_template_constraint(candidate))

        for constraint in constraints:
            if hasattr(constraint, 'reset'):
                constraint.reset() 

        candidate.set_constraint(constraints)

    def get_template_constraint(self, candidate):
        return FixAtoms(indices=np.arange(len(candidate.template)))



from agox.postprocessors.ABC_postprocess import PostprocessBaseClass

class WrapperPostprocess(PostprocessBaseClass):

    """
    Wraps a candidate object such that all atoms are in the  first unit cell.

    Parameters
    ----------
    None    
    """

    name = 'WrapperTheRapper'

    @PostprocessBaseClass.immunity_decorator
    def postprocess(self, candidate):
        candidate.wrap()
        return candidate
        


import numpy as np
from scipy.linalg import svd

from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass


class OldCUR(SparsifierBaseClass):
    name = "OldCUR"

    def __init__(self, m_points: int = 1000, **kwargs) -> None:
        super().__init__(m_points, **kwargs)
        DeprecationWarning("This is an older incorrect implementation of CUR, only kept for comparison purposes. Use CUR instead.")

    def sparsify(self, X: np.ndarray) -> np.ndarray:
        if X.shape[0] < self.m_points:
            m_indices = np.arange(0, X.shape[0])
            return X, m_indices

        U, _, _ = svd(X)
        score = np.sum(U[:, : self.m_points] ** 2, axis=1) / self.m_points
        sorter = np.argsort(score)[::-1]
        Xm = X[sorter, :][: self.m_points, :]

        return Xm, sorter[: self.m_points]



"""
Many algorithms use generators that are seeded by previous candidates, e.g. in 
Basin-Hopping the next candidate is generated by a small random perturbation of
the previous candidate.

Samplers are responsible for keeping track of which candidates to use as seeds
for the next generation of candidates. 
"""
from .ABC_sampler import SamplerBaseClass
from .spectral_graph import SpectralGraphSampler
from .kmeans import KMeansSampler
from .metropolis import MetropolisSampler
from .concurrent_tempering import ConcurrentTemperingSampler
from .genetic import GeneticSampler, DistanceComparator
from .kernel_similarity import KernelSimSampler
from .fixed_sampler import FixedSampler
from .parallel_tempering import ParallelTemperingSampler

__all__ = ['SamplerBaseClass',
            'SpectralGraphSampler',
            'KMeansSampler',
            'MetropolisSampler',
            'ConcurrentTemperingSampler',
            'ParallelTemperingSampler',
            'GeneticSampler',
            'DistanceComparator',
            'KernelSimSampler',
            'FixedSampler',
        ]


import ipywidgets as widgets # New
import matplotlib.pyplot as plt
import numpy as np
from datetime import date
import os

from agox.utils.plot import plot_atoms, plot_cell

FONTCOLOR = 'black'

def make_save_folder():
    today = date.today().strftime('%d_%m_%y')
    folder = os.path.expanduser('~')+'/figs/'+today + '/'

    if not os.path.exists(folder):
        os.mkdir(folder)
    return folder

def savefig(path, fig):
    if not os.path.exists(path):
        fig.savefig(path, bbox_inches='tight')
    else:
        print('Warning: Figure with that name already exists')

####################################################################################################################
# Interactive: Functions for interactive Jupyter notebook plots
####################################################################################################################

class InteractiveStructureHistogram:

    def __init__(self, analysis, xlim=None, ylim=None, template_indices=None, histogram_kwargs=None):

        # Attached 'Analysis' object:
        self.analysis = analysis

        # Get all best structures and their energies:
        self.structures, self.energies = self.analysis.get_best_structures()

        self.first_call = True

        self.xlim = xlim
        self.ylim = ylim

        self.histogram_kwargs = histogram_kwargs
        self.template_indices = template_indices

    def update_plot(self, index):
        fig, ax = plt.subplots(1, 2, figsize=(14, 5))

        #self.analysis.plot_structure_nice(ax[0], self.structures[index])
        if self.template_indices is None:
            plot_atoms(ax[0], self.structures[index], plot_constraint=True, repeat=1)
        else:
            structure_indices = [i for i in range(len(self.structures[index])) if i not in self.template_indices]
            plot_atoms(ax[0], self.structures[index][self.template_indices], plot_constraint=True, patch_kwargs=dict(alpha=0.5), repeat=1)
            plot_atoms(ax[0], self.structures[index][structure_indices], plot_constraint=True, repeat=1)

        plot_cell(ax[0], self.structures[index].cell)

        ax[0].set_title('Structure {}: Energy = {:7.2f}'.format(index, self.energies[index]))

        # Histogram
        if self.first_call:
            self.analysis.plot_histogram(ax[1], self.histogram_kwargs)
            limits = ax[1].get_ylim()
            self.hist_line, = ax[1].plot([self.energies[index], self.energies[index]], [0, limits[1]], color='black')
        else:
            self.hist_line.set_xdata([self.energies[index], self.energies[index]])

        #self.first_call = False

class InteractiveSuccessStats:

    def __init__(self, analysis):

        self.analysis = analysis

        self.CDF = self.analysis.CDF # Convenience/Lazyness
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

        self.label_fontsize = self.analysis.label_fontsize
        self.tick_sizes = self.analysis.tick_sizes

        self.where_step = 'post'
        self.figsize = (6, 6)

        # Widgets:
        self.selector = widgets.SelectMultiple(options=self.analysis.labels, value=self.analysis.labels, description='Datasets')
        self.xmin = widgets.BoundedIntText(value=0, min=0, max=self.analysis.max_num_iterations, step=1, description='Min iterations:')
        self.xmax = widgets.BoundedIntText(value=self.analysis.max_num_iterations, min=0, max=self.analysis.max_num_iterations, step=1, description='Max iterations:')

        self.uncertainty_button = widgets.ToggleButtons(options={'On':True, 'Off':False}, 
                description='Uncertainty', 
                button_style='', 
                tooltips=['Show uncertainty', 'Hide uncertainty'], 
                disabled=False)

        self.label_button = widgets.ToggleButtons(
                options={'On':True, 'Off':False}, 
                description='Labels', 
                button_style='', 
                tooltips=['Show labels', 'Hide labes'], 
                disabled=False)               

        self.rate_button = widgets.ToggleButtons(
                options={'On':True, 'Off':False}, 
                description='Rate', 
                button_style='', 
                tooltips=['Show rate', 'Hide rate'], 
                disabled=False, 
                value=False)

        #self.rate_button = widgets.ToggleButton(value=False, description='Ep. for rate', button_style='', tooltip='Control whether or not iteration for rate is plotted', icon='check')
        self.rate_input = widgets.BoundedFloatText(value=0.5, min=0, max=1, step=0.01, description='Rate:')

        # Save button:        
        # self.save_button = widgets.ToggleButton(value=False, description='Save figure?', button_style='danger')
        # self.save_name = widgets.Text(value='default', description='Figure save name')

        self.widget_dict = {'labels':self.selector, 'xmin':self.xmin, 'xmax':self.xmax, 'uncertainty':self.uncertainty_button, 
                            'plot_labels':self.label_button, 'plot_iteration_for_rate':self.rate_button, 'rate':self.rate_input}
        self.widget_list = [self.widget_dict[key] for key in self.widget_dict.keys()]

    def plot_succes_curve(self, ax, index, uncertainty=True):
        color = self.colors[index % len(self.colors)]
        where_step = 'post'
        if uncertainty:
            try:
                ax.fill_between(self.CDF[index][0], self.CDF[index][2], self.CDF[index][3], step=where_step, facecolor=color, alpha=0.1)
            except:
                ax.fill_between(self.CDF[index][0], self.CDF[index][2], self.CDF[index][3], facecolor=color, alpha=0.1)
        ax.step(self.CDF[index][0], self.CDF[index][1], where=self.where_step, c=color, label=self.analysis.labels[index])

        i = index
        max_iterations = np.max(self.analysis.iterations[i])
        if not (self.analysis.iterations[i] == max_iterations).all():
            idx = np.argmin(np.abs(
                np.array(self.CDF[i][0]).reshape(-1, 1) - np.array(self.analysis.iterations[i]).reshape(1, -1)
                ), axis=0)
            x = self.CDF[i][0][idx]
            x[x == self.CDF[i][0][-2]] = np.array(self.analysis.iterations[i])[x == self.CDF[i][0][-2]]
            ax.plot(x, self.CDF[i][1][idx], 'x', color=color)


        #max_iterations = np.max(self.analysis.iterations[index])
        #if not (self.analysis.iterations[index] == max_iterations).all():
        #    idx = np.argmin(np.abs(np.array(self.CDF[index][0]).reshape(-1, 1)-np.array(self.analysis.iterations[index]).reshape(1, -1)), axis=0)
        #    x = self.CDF[index][0][idx]
        #    x[x == self.CDF[index][0][-2]] = np.array(self.analysis.iterations[index])[x == self.CDF[index][0][-2]]
        #    ax.plot(x, self.CDF[index][1][idx], 'x', color=color)

    def plot_rate(self, ax, iteration, rate, index):
        color = self.colors[index % len(self.colors)]
        ax.plot([iteration, iteration], [0, rate], '--', color=color)

    def update_plot(self, labels, xmin, xmax, uncertainty=True, plot_labels=True, plot_iteration_for_rate=False, rate=0.5, save=False, save_name=None):
        fig, ax = plt.subplots(figsize=self.figsize)
        
        if plot_iteration_for_rate:
            iteration_for_rate = self.analysis.get_iterations_for_rate(rate)

        for i, label in enumerate(self.analysis.labels):
            if label in labels:
                self.plot_succes_curve(ax, index=i, uncertainty=uncertainty)

                if plot_iteration_for_rate:
                    self.plot_rate(ax, iteration_for_rate[i], rate, i)

        ax.set_xlabel('Iterations [#]', fontsize=self.label_fontsize, color=FONTCOLOR)
        ax.set_ylabel('Success [%]', fontsize=self.label_fontsize,  color=FONTCOLOR)
        ax.set_ylim([0, 1])
        ax.set_xlim([xmin, xmax])
        ax.tick_params(axis='both', which='major', labelsize=self.tick_sizes, labelcolor=FONTCOLOR)
        if plot_labels:
            ax.legend(fontsize=self.label_fontsize, loc='best')

        if save:
            folder = make_save_folder()
            savefig(folder + save_name + '.png', fig)
            self.save_button.value = False

class InteractiveEnergy:

    def __init__(self, analysis):

        self.analysis = analysis

        self.figsize = (6, 6)
        self.label_fontsize = self.analysis.label_fontsize
        self.tick_sizes = self.analysis.tick_sizes
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']


        # Widgets:
        self.selector = widgets.SelectMultiple(options=self.analysis.labels, value=self.analysis.labels, description='Datasets')
        self.xmin = widgets.BoundedIntText(value=0, min=0, max=self.analysis.max_num_iterations, step=1, description='Min iterations:')
        self.xmax = widgets.BoundedIntText(value=self.analysis.max_num_iterations, min=0, max=self.analysis.max_num_iterations, step=1, description='Max iterations:')
        self.ymin_offset = widgets.BoundedFloatText(value=1, min=0, max=25, description='Min. energy')
        self.ymax_offset = widgets.BoundedFloatText(value=50, min=0, max=1000, description='Max. energy')
        
        self.mean_button = widgets.ToggleButtons(
            options={'On':True, 'Off':False},
            value=True, 
            description='Mean energy', 
            button_style='', 
            tooltips=['Show mean', 'Hide mean'], 
            icon='check')

        self.limit_button = widgets.ToggleButtons(
            options={'On':True, 'Off':False},
            value=False, 
            description='Energy limits', 
            button_style='', 
            tooltips=['Show limits', 'Hide limits'], 
            icon='check')

        self.rolling_average = widgets.BoundedIntText(value=1, min=1, max=500, description='Rolling average #:')
        
        # # Save button:        
        # self.save_button = widgets.ToggleButton(value=False, description='Save figure?', button_style='danger')
        # self.save_name = widgets.Text(value='default', description='Figure save name')

        # Widget dict/list
        self.widget_dict = {'labels':self.selector, 'xmin':self.xmin, 'xmax':self.xmax, 'ymin':self.ymin_offset, 'ymax':self.ymax_offset,
                            'plot_mean':self.mean_button, 'plot_limit':self.limit_button, 'rolling_average':self.rolling_average}
        self.widget_list = [self.widget_dict[key] for key in self.widget_dict.keys()]

    def update_plot(self, labels, xmin, xmax, ymin, ymax, plot_mean, plot_limit, rolling_average):
        fig, ax = plt.subplots(figsize=self.figsize)

        energies = self.analysis.energies
        iterations = np.arange(energies.shape[-1])

        policy_only = False

        if policy_only:
            iterations = np.arange(5, energies.shape[-1], 5)
            energies = energies[:, :, iterations-1]
            
        for index, label in enumerate(self.analysis.labels):
            
            if label in labels:

                self.plot_best_energy(ax, index)

                if plot_mean:
                    if not policy_only:
                        self.plot_mean_energy(ax, index)
                    self.plot_mean_energy(ax, index, policy=True)
                    
                if plot_limit:

                    if policy_only:
                        self.plot_energy_limits(ax, index, rolling_average, policy=True)
                    else:
                        self.plot_energy_limits(ax, index, rolling_average, policy=False)

        ax.plot([xmin, xmax],[self.analysis.global_best_energy, self.analysis.global_best_energy], '--k')
        ax.legend()
        ax.set_ylim([self.analysis.global_best_energy-ymin, self.analysis.global_best_energy+ymax])
        ax.set_xlim([xmin, xmax])

        ax.set_xlabel('iteration [#]', fontsize=self.label_fontsize)
        ax.set_ylabel('Energy [eV]', fontsize=self.label_fontsize)

    def plot_best_energy(self, ax, index):
        ax.plot(np.mean(self.analysis.best_energies[index, 0:self.analysis.restarts[index], :], axis=0), '-', label=self.analysis.labels[index],
                    color=self.colors[index])
    
    def plot_mean_energy(self, ax, index, policy=False):
        energies = self.analysis.energies
        iterations = np.arange(energies.shape[-1])
        alpha = 0.8
        color = self.colors[index]
        if policy:
            iterations = np.arange(5, energies.shape[-1], 5)
            energies = energies[:, :, iterations-1]
            alpha = 0.8

        arr = np.nanmean(energies[index, 0:self.analysis.restarts[index], :], axis=0)

        if policy:
            ax.plot(iterations, arr, linestyle='-', label='_nolegend_',
                color='black', alpha=0.5)

        ax.plot(iterations, arr, linestyle='-', label='_nolegend_',
                    color=color, alpha=alpha)
    
    def plot_energy_limits(self, ax, index, w, policy=False):
        energies = self.analysis.energies
        iterations = np.arange(energies.shape[-1])

        if policy:
            iterations = np.arange(5, energies.shape[-1], 5)
            energies = energies[:, :, iterations-1]

        upper_limit = np.convolve(np.max(energies[index, 0:self.analysis.restarts[index]], axis=0), np.ones(w), 'same') / w
        lower_limit = np.convolve(np.min(energies[index, 0:self.analysis.restarts[index]], axis=0), np.ones(w), 'same') / w

        ax.fill_between(iterations, lower_limit, upper_limit, color=self.colors[index], alpha=0.4)

from ase import Atoms
def sorted_species(atoms):
    sorted_atoms = Atoms('', cell=atoms.get_cell())

    indices = np.argsort(atoms.get_atomic_numbers())
    for idx in indices:
        sorted_atoms += atoms[idx]

    return sorted_atoms



import ray
import numpy as np

from typing import Callable, List



@ray.remote(num_cpus=1, runtime_env={'env_vars': {
    'OMP_NUM_THREADS': '1',
    'OPENBLAS_NUM_THREADS': '1',
    'MKL_NUM_THREADS': '1',
    }}, 
    max_restarts=5)
class Actor:

    """
    Basic Ray Actor used by AGOX.
    """

    def __init__(self, actor_id: int = 0):
        self.modules = {}
        self.id = actor_id

    def execute_function(self, fn: Callable, module_keys: List, *args, **kwargs):
        """
        Execute a function on the actor.

        Parameters:
        -----------
        fn: function
            Function to execute.
        module_keys: list
            List of keys to modules to pass to the function.
        args: list
            Arguments to pass to the function.
        kwargs: dict
            Keyword arguments to pass to the function.
        """
        kwargs = self.set_seed(**kwargs)
        return fn(*[self.modules[key] for key in module_keys], *args, **kwargs)

    def set_seed(self, **kwargs):
        """
        Set the seed for the actor.
        """
        seed = kwargs.pop("pool_internal_seed", None)
        if seed is not None:
            np.random.seed(seed)
        return kwargs

    def add_module(self, module, key):
        """
        Add a module to the actor.

        Parameters:
        -----------
        module: AGOX module
            The module to add to the actor.
        key: module.ray_key (uuid).
            The key that references the module on the actors.
        """
        self.modules[key] = module

    def remove_module(self, key):
        """
        Remove a module from the actor.
        """
        del self.modules[key]



from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase import Atom
import numpy as np

class SteepestDescentGenerator(GeneratorBaseClass):

    """
    Performs a steepest descent step on a seed structure.

    TODO: Clean-up this code. Its bad.

    Parameters
    -----------

    use_xy_only : bool
        If True, only move atoms in xy-plane.
    immunize : bool
        If True, the candidate will not be postprocessed.
    replace : bool
        ?        
    """

    name = 'SteepestDescentGenerator'

    def __init__(self, use_xy_only=False, immunize=False, replace=True,**kwargs):
        super().__init__(replace=replace,**kwargs)
        self.use_xy_only = use_xy_only
        self.immunize = immunize

    def _get_candidates(self, candidate, parents, environment):
        
        parents[0].copy_calculator_to(candidate)

        # If the Sampler is not able to give a structure with forces
        if candidate is None or candidate.calc is None or 'forces' not in candidate.calc.results:
            return []

        template = candidate.get_template()
        len_of_template = len(template)

        delta = np.random.uniform(0, 0.1)

        self.writer(self.name,': got a candidate with a force. will take step:',delta)

        candidate.positions[len_of_template:] += delta * candidate.get_forces()[len_of_template:]
        
        if self.immunize:
            candidate.set_postprocess_immunity(True)

        return [candidate]

    def get_number_of_parents(self, sampler):
        return 1
    


from agox.test.run_tests.run_utils import agox_test_run
import pytest

@pytest.mark.ray
def test_ce_gofee(tmp_path, cmd_options):
    mode = 'ce_default_gofee' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)


from agox.generators.ABC_generator import GeneratorBaseClass
from ase.data import covalent_radii
from ase.geometry import get_distances
from ase.io import write
import random
import numpy as np

from agox.generators.symmetry import SymmetryGenerator
from agox.generators.symmetry_utils.symmetry_enforcer import SymmetryEnforcer


class SymmetryPermutationGenerator(GeneratorBaseClass):
    """
    Swaps atoms in a structure while enforcing symmetry.

    Parameters
    -----------
    max_number_of_swaps : int
        Maximum number of permutations to perform.
    rattle_strength : float
        Maximum distance to permute atoms.
    use_xy_only : bool
        If True, only permute in xy-plane.
    ignore_H : bool
        If True, ignore hydrogen atoms.
    write_candidates_to_disk : bool
        If True, write candidates to disk.
    """

    name = "PermutationGenerator"

    def __init__(
        self,
        max_number_of_swaps=1,
        rattle_strength=0.0,
        use_xy_only=False,
        ignore_H=False,
        write_candidates_to_disk=False,
        replace=True,
        **kwargs,
    ):
        super().__init__(replace=replace, **kwargs)
        self.max_number_of_swaps = max_number_of_swaps
        self.rattle_strength = rattle_strength
        self.ignore_H = ignore_H
        self.use_xy_only = use_xy_only
        self.write_candidates_to_disk = write_candidates_to_disk

        if self.verbose:
            self.counter = -1

    def _get_candidates(self, candidate, parents, environment):
        if self.verbose:
            self.counter += 1

        if candidate is None:
            return []
        num = random.randrange(0, 1001, 2)

        if self.verbose and self.write_candidates_to_disk:
            write(f"candidate_{self.counter}.traj", candidate)
        number_of_atoms = len(candidate)
        number_of_template_atoms = len(candidate.get_template())
        number_of_non_template_atoms = number_of_atoms - number_of_template_atoms
        try:
            template = candidate.get_template()
            self.slab = template
            self.se = SymmetryEnforcer(
                self.slab,
                confinement_cell=self.confinement_cell,
                cell_corner=self.confinement_corner,
            )
            group_met = candidate.get_meta_information("space_group")
            table_mult = list(candidate.get_meta_information("table_mult"))
            not_sym = candidate.get_meta_information("non_sym")
            already_lost = candidate.get_meta_information("lost_sym")
            self.se.force_group = group_met
            group_met = self.se.select_group(mode=mode)
            self.not_sym = not_sym
        except:
            group_met = {
                "name": "p1",
                "system": ["rectangular", "hexagonal", "square", "oblique", "rhombic"],
                "number_max": 1,
                "axes": None,
                "rotations": None,
                "points": None,
            }
            table_mult = [1] * number_of_non_template_atoms
        template = candidate.get_template()
        if table_mult is not None:
            table = np.concatenate(([0], table_mult[:-1]))
            for i in range(len(table)):
                if i > 0:
                    table[i] = table[i] + table[i - 1]
            table += len(template)
            indices_to_permute = table

        symbols = candidate.get_atomic_numbers()[number_of_template_atoms:]
        unique_symbols = np.unique(symbols)
        if self.ignore_H:
            unique_symbols = np.array([s for s in unique_symbols if s != 1])

        number_of_unique_symbols = len(unique_symbols)
        assert (
            number_of_unique_symbols > 1
        ), "Cannot be used for single component systems"
        write("permute.vasp", candidate)

        number_of_swaps = np.random.randint(self.max_number_of_swaps) + 1

        for n in range(number_of_swaps):
            symbol_i = unique_symbols[np.random.randint(number_of_unique_symbols)]
            remaining_symbols = np.delete(
                unique_symbols,
                [
                    idx
                    for idx in range(number_of_unique_symbols)
                    if unique_symbols[idx] == symbol_i
                ],
            )

            symbol_j = remaining_symbols[
                np.random.randint(number_of_unique_symbols - 1)
            ]

            idx_symbol_i = (
                np.argwhere(symbols == symbol_i).reshape(-1) + number_of_template_atoms
            )
            idx_symbol_j = (
                np.argwhere(symbols == symbol_j).reshape(-1) + number_of_template_atoms
            )

            # can only permute the first index of each table_mult
            idx_symbol_i = intersection = np.intersect1d(
                indices_to_permute, idx_symbol_i
            )
            idx_symbol_j = intersection = np.intersect1d(
                indices_to_permute, idx_symbol_j
            )

            combinations_ij = np.array(
                np.meshgrid(idx_symbol_i, idx_symbol_j)
            ).T.reshape(-1, 2)

            for row in np.random.permutation(combinations_ij):
                swap_idx_i = row[0]
                swap_idx_j = row[1]
                # here the combination should only happend if multiplicity is the same for both atoms
                tab1 = int(np.argwhere(indices_to_permute == swap_idx_i))
                tab2 = int(np.argwhere(indices_to_permute == swap_idx_j))

                if table_mult[tab1] != table_mult[tab2]:
                    continue
                new_positions = candidate.get_positions()
                for sym_t in range(table_mult[tab1]):
                    try:
                        new_positions[
                            [swap_idx_i + sym_t, swap_idx_j + sym_t]
                        ] = new_positions[[swap_idx_j + sym_t, swap_idx_i + sym_t]]
                    except:
                        return []
                """
                if self.use_xy_only:
                    new_positions[swap_idx_i] +=self.pos_add_disk(self.rattle_strength)
                    new_positions[swap_idx_j] +=self.pos_add_disk(self.rattle_strength)
                else:
                    new_positions[swap_idx_i] +=self.pos_add_sphere(self.rattle_strength)
                    new_positions[swap_idx_j] +=self.pos_add_sphere(self.rattle_strength)
                """

                swap_successfull = True
                near_enough_to_other_atoms = False
                for i in (swap_idx_i, swap_idx_j):
                    for other_atom_idx in range(len(candidate)):
                        if other_atom_idx in [swap_idx_i, swap_idx_j]:
                            continue
                        other_atom = candidate[other_atom_idx]

                        covalent_dist = (
                            covalent_radii[candidate[i].number]
                            + covalent_radii[other_atom.number]
                        )

                        rmin = 0.85 * covalent_dist
                        rmax = 1.15 * covalent_dist
                        tmp = np.linalg.norm(other_atom.position - new_positions[i])
                        if (
                            np.linalg.norm(other_atom.position - new_positions[i])
                            < rmin
                        ):
                            swap_successfull = False
                            break
                        if (
                            np.linalg.norm(other_atom.position - new_positions[i])
                            < rmax
                        ):
                            near_enough_to_other_atoms = True
                    if not swap_successfull or not near_enough_to_other_atoms:
                        break

                if not swap_successfull or not near_enough_to_other_atoms:
                    continue

                candidate.set_positions(new_positions)
                if self.verbose and self.write_candidates_to_disk:
                    write(f"candidate_swap_{n}_{self.counter}.traj", candidate)
                break
            else:
                self.writer("No swaps possible")
                return []

        return [candidate]

    def pos_add_disk(self, rattle_strength):
        """Help function for rattling within a disk"""
        r = rattle_strength * np.random.rand() ** (1 / 2)
        theta = np.random.uniform(low=0, high=2 * np.pi)
        pos_add = r * np.array([np.cos(theta), np.sin(theta), 0])
        return pos_add

    def pos_add_sphere(self, rattle_strength):
        """Help function for rattling within a sphere"""
        r = rattle_strength * np.random.rand() ** (1 / 3)
        theta = np.random.uniform(low=0, high=2 * np.pi)
        phi = np.random.uniform(low=0, high=np.pi)
        pos_add = r * np.array(
            [np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)]
        )
        return pos_add

    def get_number_of_parents(self, sampler):
        return 1



import jax
import numpy as np
import jax.numpy as jnp

from matscipy.neighbours import neighbour_list, triplet_list
from matscipy.ffi import first_neighbours

from ase import Atoms

from typing import Dict, Optional
from functools import partial

from dataclasses import dataclass


@dataclass
class AtomsData:
    R: jnp.ndarray
    Z: jnp.ndarray
    idx_i: jnp.ndarray
    idx_j: jnp.ndarray
    cell_shift: jnp.ndarray
    cell: jnp.ndarray
    idx_triplet_ij: Optional[jnp.ndarray] = None
    idx_triplet_ik: Optional[jnp.ndarray] = None


class AtomsConverter:

    def __init__(self, r_cut: float = 5.0,
        collect_triplet: bool = True,
        r_cut_angular: float = 5.0,
        volume_min: float = 1e-5,
        max_pair_size: int = 0,
        max_triplet_size: int = 0,
        pad: bool = False):

        self.r_cut = r_cut
        self.collect_triplet = collect_triplet
        self.r_cut_angular = r_cut_angular
        self.volume_min = volume_min
        self.max_pair_size = max_pair_size
        self.max_triplet_size = max_triplet_size
        self.pad = pad

    def convert(self, atoms: Atoms) -> Dict[str, np.ndarray]:
        """
        Convert an ASE Atoms object to ML inputs

        Parameters:
            atoms: an ASE Atoms object
            r_cut: float, radius cutoff for determining neighbors
            collect_triplet: boolean, whether indices of triplets are collected
            r_cut_angular: float, angular cutoff for determining triplets
            n_max_neighbors: int, maintain a constant shape of neighbour_list
            volume_min: float, minimum volume of atoms (matscipy needs volume in neighbour_list)
        """

        R = atoms.get_positions()
        Z = atoms.get_atomic_numbers()


        # avoid too small cell in matscipy neighborlist calculation
        # useful for non-periodic structure and when atoms.cell is None
        if atoms.cell.volume < self.volume_min:
            atoms = atoms.copy()
            cell_buffer = 2 * self.r_cut
            new_cell = np.ptp(atoms.get_positions(), axis=0) + cell_buffer
            atoms.set_cell(new_cell)

        # each pair will be generated twice
        idx_i, idx_j, r_ij, cell_shift = neighbour_list("ijdS", atoms, cutoff=self.r_cut)
        n_pair_raw = len(idx_i)
        if self.pad:

            max_pair_size = np.max([self.max_pair_size, len(idx_i)])
            n_pair_pad = max_pair_size-len(idx_i)
            pad_mode = 'constant'

            idx_i = np.pad(idx_i, (0, n_pair_pad), pad_mode)
            idx_j = np.pad(idx_j, (0, n_pair_pad), pad_mode)
            
            # this makes sure that r_ij at padded indices are larger than the set cutoff
            #   therefore zero contribution to feature values
            cell_shift = np.pad(cell_shift, ((0, n_pair_pad), (0, 0)), pad_mode, constant_values=10)

        inputs = {
            "R": R,
            "Z": Z,
            "idx_i": idx_i,
            "idx_j": idx_j,
            "cell_shift": cell_shift,
            "cell": atoms.get_cell()[:],  # do not store ASE Cell object
        }

        if self.collect_triplet:
            first_n = first_neighbours(len(atoms), idx_i[:n_pair_raw])
            ij_triplet, ik_triplet = triplet_list(first_n, r_ij, self.r_cut_angular)

            if self.pad:
                max_triplet_size = np.max([self.max_triplet_size, len(ij_triplet)])
                n_triplet_pad = max_triplet_size - len(ij_triplet)
                ij_triplet = np.pad(ij_triplet, (0, n_triplet_pad), pad_mode)
                ik_triplet = np.pad(ik_triplet, (0, n_triplet_pad), pad_mode)
            
            inputs["idx_triplet_ij"] = ij_triplet
            inputs["idx_triplet_ik"] = ik_triplet

        data = AtomsData(**inputs)
        self.max_pair_size = max_pair_size
        self.max_triplet_size = max_triplet_size
        
        return data


def radial_cutoff(r: float, r_cut: float) -> float:
    """
    A hard cutoff function
    """
    return r * (r < r_cut)


def angular_cutoff(r: float, r_cut: float, gamma: int = 2) -> float:
    """
    A cutoff function for angular features
    """
    value = 1 + gamma * (r / r_cut) ** (gamma + 1) - (gamma + 1) * (r / r_cut) ** gamma
    value += gamma == 0
    value *= r < r_cut
    return value


def angle_between(v1, v2):
    """Returns the angle in radians between two unit vectors 'v1' and 'v2'
    https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python
    """
    return jnp.arccos(jnp.clip(jnp.dot(v1, v2), -1.0, 1.0))


@partial(jax.vmap, in_axes=(0, 0))
def element_wise_angle_between(*args):
    return angle_between(*args)


def gaussian_cdf(x: float, mu: float, sigma: float) -> float:
    """
    Calculate gaussian cumulative distribution function (cdf) from lower bound -inf to upper bound `x`

    Parameters:
        x: upper bound of the cdf
        mu: center of the gaussian
        sigma: standard deviation of the gaussian
    """
    cdf_x = 0.5 * (1 + jax.scipy.special.erf((x - mu) / sigma / jnp.sqrt(2)))
    return cdf_x


@partial(jax.vmap, in_axes=(None, 0, None), out_axes=0)
def gaussian_cdf_matrix(*args):
    """
    return: shape (n_mu, n_x)
    """
    return gaussian_cdf(*args)


@partial(jax.jit, static_argnums=[6, 7, 8, 9, 10])
def compute_radial_features(
    R: jax.Array,
    idx_i: jax.Array,
    idx_j: jax.Array,
    idx_pair: jax.Array,
    cell_shift: jax.Array,
    cell: jax.Array,
    n_unique_pair: int,  # static_arg from here
    r_cut: float,
    n_bins: int,
    sigma: float,
    prefactor: float,
) -> jax.Array:
    """
    Compute radial features

    Parameters:
        R: atomic positions (n_atoms, 3)
        idx_i: atomic index of atom 'i' in the pair (n_pair, )
        idx_j: atomic index of atom 'j' in the pair (n_pair, )
        idx_pair: pair index (n_pair, )
        cell_shift: cell_shift
        cell: cell
        n_unique_pair: number of unique pair
        r_cut: radial cutoff
        n_bins: number of bins
        sigma: gaussian width
        prefactor: prefactor on feature values

    Return:
        Radial features, (n_unique_pair*n_bins, )
    """
    R_ij = R[idx_i] - R[idx_j] + cell_shift.dot(cell)
    r_ij = jnp.linalg.norm(R_ij, axis=1, keepdims=True)

    x = jnp.linspace(0, r_cut, n_bins + 1)

    cdf = gaussian_cdf_matrix(x, r_ij, sigma)

    # AGOX's Cython implementation does not have this prefactor
    cdf *= sigma * jnp.sqrt(2 * jnp.pi)
    cdf *= 1 / r_ij**2

    integral_in_bins = cdf[:, 1:] - cdf[:, :-1]

    radial_features = jnp.zeros((n_unique_pair, n_bins))
    radial_features = radial_features.at[idx_pair].add(integral_in_bins)
    radial_features = 1 * radial_features.flatten()  # avoid double sum i,j and j,i

    radial_features *= prefactor

    return radial_features


@partial(jax.jit, static_argnums=[6, 7, 8, 9, 10])
@jax.jacfwd
def compute_radial_feature_grad(*args):
    """
    Compute the derivative of radial features w.r.t the first argument (atomic positions)

    Parameters are similar with compute_radial_features()

    Return:
        Rarial feature gradients, (n_unique_pair*n_bins, n_atoms, 3)
    """
    # print('radial grad')
    return compute_radial_features(*args)


@partial(jax.jit, static_argnums=[8, 9, 10, 11, 12, 13])
def compute_angular_features(
    R: jax.Array,
    idx_i: jax.Array,
    idx_j: jax.Array,
    idx_triplet_ij: jax.Array,
    idx_triplet_jk: jax.Array,
    idx_triplet: jax.Array,
    cell_shift: jax.Array,
    cell: jax.Array,
    n_unique_triplet: int,  # static_arg from here
    r_cut: float,
    n_bins: int,
    sigma: float,
    gamma: int,
    prefactor: float,
) -> jax.Array:
    """
    Compute angular features

    Parameters are similar with compute_radial_features()

    Return:
        Angular features, (n_unique_triplet*n_bins, )
    """
    R_ij = R[idx_i] - R[idx_j] + cell_shift.dot(cell)
    r_ij = jnp.linalg.norm(R_ij, axis=1, keepdims=True)

    R_ij_triplet = R_ij[idx_triplet_ij]
    R_ik_triplet = R_ij[idx_triplet_jk]

    r_ij_triplet = r_ij[idx_triplet_ij]
    r_ik_triplet = r_ij[idx_triplet_jk]

    # angles centered at atom 'i' for all angles
    angle_ijk = element_wise_angle_between(
        R_ij_triplet / r_ij_triplet, R_ik_triplet / r_ik_triplet
    )

    x = jnp.linspace(0, jnp.pi, n_bins + 1)
    cdf = gaussian_cdf_matrix(x, angle_ijk, sigma)
    integral_in_bins = cdf[:, 1:] - cdf[:, :-1]

    integral_in_bins *= sigma * jnp.sqrt(2 * jnp.pi)
    integral_in_bins *= angular_cutoff(r_ij_triplet, r_cut, gamma) * angular_cutoff(
        r_ik_triplet, r_cut, gamma
    )
    # make sure padded triplets have zero contributions to angular features
    integral_in_bins *= jnp.where(idx_triplet_ij == idx_triplet_jk, 0, 1)[:, None]

    angular_features = jnp.zeros((n_unique_triplet, n_bins))
    angular_features = angular_features.at[idx_triplet].add(integral_in_bins)
    angular_features = (
        0.5 * angular_features.flatten()
    )  # avoid double sum i,j,k and i,k,j

    angular_features *= prefactor

    return angular_features


@partial(jax.jit, static_argnums=[8, 9, 10, 11, 12, 13, 14])
@jax.jacfwd
def compute_angular_feature_grad(*args):
    """
    Compute the derivative of angular features w.r.t the first argument (atomic positions)

    Parameters are similar with compute_angular_features()

    Return:
        Angular feature gradients, (n_unique_triplet*n_bins, n_atoms, 3)
    """
    # print('angular grad')
    return compute_angular_features(*args)


@jax.jit
def get_pair_index(
    unique_pairs: jax.Array, z_i: jax.Array, z_j: jax.Array
) -> jax.Array:
    """
    Get the sorted index of a given pair

    Parameters:
        unique_pairs: jax.Array, atomic numbers of unique pairs
        z_i: jax.Array, atomic number of atom 'i' in the pair
        z_j: jax.Array, atomic number of atom 'j' in the pair

    Return:
        jax.Array, shape: (n_pair, )
    """
    z_ij = jnp.sum(
        jnp.stack((z_i, z_j), axis=1).sort(axis=1) * jnp.array([10, 1]), axis=1
    )
    idx_pair = unique_pairs.searchsorted(z_ij)
    return idx_pair


@jax.jit
def get_triplet_index(
    unique_triplets: jax.Array, z_i: jax.Array, z_j: jax.Array, z_k: jax.Array
) -> jax.Array:
    """
    Get the sorted index of a given triplet

    Parameters:
        unique_triplets: jax.Array, atomic numbers of unique triplets
        z_i: jax.Array, atomic number of atom 'i' in the triplet
        z_j: jax.Array, atomic number of atom 'j' in the triplet
        z_k: jax.Array, atomic number of atom 'k' in the triplet
    Return:
        jax.Array, shape: (n_triplet, )
    """
    # assert len(z_i) == len(z_j) == len(z_k)
    z_jk_sorted = jnp.sum(
        jnp.stack((z_j, z_k), axis=1).sort(axis=1) * jnp.array([10, 1]), axis=1
    )
    z_ijk = 100 * z_i + z_jk_sorted
    idx_triplet = unique_triplets.searchsorted(z_ijk)
    return idx_triplet



def add_analysis_parser(subparsers):
    # Input arguments:

    parser = subparsers.add_parser("analysis", help="Analysis of search results.")
    parser.set_defaults(func=cli_analysis)    

    parser.add_argument("-d", "--directories", nargs="+", type=str)  # List - Directories
    parser.add_argument("-dE", "--delta_energy", type=float, default=0.05)
    parser.add_argument("-DE", "--Delta_energy", type=float, default=None)
    parser.add_argument("--save_name", type=str, default=None)

    # Flags
    parser.add_argument("-c", "--plot_cdf", action="store_false")  # Bool - Plot CDF
    parser.add_argument("-s", "--plot_structure", action="store_false")  # Bool - Plot structure
    parser.add_argument("-e", "--plot_energies", action="store_true")  # Bool - Plot energy
    parser.add_argument("-hg", "--plot_histogram", action="store_false")  # Bool - Plot histogram
    parser.add_argument("-p", "--print_all", action="store_true")  # Print everything
    parser.add_argument("-v", "--view_best", action="store_true")
    parser.add_argument("-g", "--save", action="store_true")  # Save the figure?
    parser.add_argument("-sd", "--save_data", action="store_false")
    parser.add_argument("-fr", "--force_reload", action="store_true")

    # Histogram settings:
    parser.add_argument(
        "-hgl", "--histogram_limits", nargs="+", default=[0, 0]
    )  # Histogram limits: min iteration, max iteration
    parser.add_argument("-hgop", "--histogram_operation", type=str, default="best")  # Histogram operations: best, all
    parser.add_argument(
        "-hgth", "--histogram_threshold", type=int, default=1000
    )  # Histogram energy threshold: 0 gives all (negative) energies, positive number gives that over the best found.

    # Animation related things:
    parser.add_argument("-a", "--animate", action="store_false")  # Print everything

def cli_analysis(args):
    from agox.utils.batch_analysis import Analysis, KeyBoardEvent
    import matplotlib.pyplot as plt

    directories = args.directories

    if args.Delta_energy is None:
        dE = args.delta_energy
        delta_energy = None
    else:
        dE = None
        delta_energy = args.Delta_energy

    # Plots:
    plot_cdf = args.plot_cdf
    plot_structure = args.plot_structure
    plot_energies = args.plot_energies
    plot_histogram = args.plot_histogram
    num_plots = plot_cdf + plot_structure + plot_energies + plot_histogram
    print_all = args.print_all
    view_best = args.view_best

    histogram_kwargs = {
        "min_hist_iteration": int(args.histogram_limits[0]),
        "max_hist_iteration": int(args.histogram_limits[1]),
        "operation": args.histogram_operation,
        "threshold": args.histogram_threshold,
    }

    # Save/reload
    save_data = args.save_data
    force_reload = args.force_reload

    # Saving?
    save = args.save
    save_name = args.save_name

    animate = args.animate

    # Start
    A = Analysis(
        directories,
        dE=dE,
        save_data=save_data,
        force_reload=force_reload,
        histogram_kwargs=histogram_kwargs,
        delta_energy=delta_energy,
    )
    # A = Analysis(directories, dE=dE, save_data=True, force_reload=False)

    A.analyse_directories()
    A.calculate_CDF()

    A.print_info(print_all)

    # Open the best structure in the ASE GUI.
    if view_best:
        A.global_best_structure.edit()

    if num_plots > 0:
        fig, axes = plt.subplots(1, num_plots, figsize=(num_plots * 5, 5))
        if num_plots > 1:
            axes = axes.flatten()
        else:
            axes = [axes]

        ax_idx = 0
        if plot_structure:
            if animate:
                atoms_patch_collection, structures, energies = A.animate_structure(axes[ax_idx])
            else:
                A.plot_structure(axes[ax_idx], A.global_best_structure)

            struct_ax = axes[ax_idx]
            ax_idx += 1
        else:
            struct_ax = None

        if plot_cdf:
            A.plot_CDF(axes[ax_idx])
            cdf_ax = axes[ax_idx]
            ax_idx += 1
        else:
            cdf_ax = None

        if plot_energies:
            A.plot_energy(axes[ax_idx])
            energy_ax = axes[ax_idx]
            ax_idx += 1
        else:
            energy_ax = None

        if plot_histogram:
            A.plot_histogram(axes[ax_idx])
            hist_ax = axes[ax_idx]
            ax_idx += 1
        else:
            hist_ax = None

        if animate:
            if plot_structure:
                named_axes = [struct_ax, cdf_ax, energy_ax, hist_ax]
                callback = KeyBoardEvent(fig, atoms_patch_collection, structures, energies, named_axes=named_axes)
                fig.canvas.mpl_connect('key_press_event', callback.press)
            else:
                print("To use the animation, the structure has to be plotted")

        plt.tight_layout()

        if save:
            if save_name is None:
                save_name = directories[0]
            plt.savefig(
                os.path.join(save_name + strftime("%Y_%m_%d") + ".png"),
                bbox_inches="tight",
            )

        plt.show()



from ase.build import molecule
import matplotlib.pyplot as plt

from agox.models.descriptors.fingerprint_jax.fingerprint_jax import JaxFingerprint
import numpy as np

def main():

    #atoms = molecule('CH3CH2OH')
    atoms = molecule('C6H6')
    atoms.center(5)

    print(atoms.get_pbc(), len(atoms), atoms.get_chemical_formula())

    use_angular = True
    eta = 5
    r_cut_all = 5.0
    n_bins = 30
    descriptor = JaxFingerprint.from_atoms(atoms, r_cut_radial=r_cut_all, r_cut_angular=r_cut_all,
                                         n_bins_radial=n_bins, use_angular=use_angular, eta=eta)

    print(descriptor.pairs)
    print(descriptor.triplets)

    features = descriptor.get_features(atoms).flatten()

    from agox.models.descriptors.fingerprint_cython.angular_fingerprintFeature_cy import Angular_Fingerprint
    descriptor_agox = Angular_Fingerprint(atoms, Rc1=r_cut_all, Rc2=r_cut_all, binwidth1=r_cut_all/n_bins, nsigma=10,
                                        sigma1=0.2, sigma2=0.2, Nbins2=n_bins,
                                        eta=eta, gamma=2, use_angular=use_angular)
    features_agox = descriptor_agox.get_feature(atoms)

    plt.switch_backend('Agg')
    _, ax = plt.subplots(figsize=(6,4))
    x = np.arange(len(features))
    y1 = 2*features
    y2 = -features_agox
    ax.bar(x, y1, width=0.5)
    ax.bar(x, y2, width=0.5)
    ax.text(0, 10, 'Jax', color='C0')
    ax.text(0, -10, 'Cython', color='C1')
    
    x_bins = np.arange(0, len(features)+1, n_bins)
    
    labels = ['HH', 'HC', 'CC', 'HHH', 'HHC', 'HCC', 'CHH', 'CHC', 'CCC']
    for i in range(len(labels)):
        x = (x_bins[i+1] + x_bins[i]) / 2
        y = -25
        ax.text(x, y, labels[i], ha='center')
    ax.set_xticks(x_bins)
    ax.set_xticklabels(x_bins)

    ax.set_xlabel('Feature index')
    ax.set_ylabel('Feature value')

    ax.set_title('Fingerprint of C6H6 molecule')
    
    
    plt.savefig('fig-fingerprint-C6H6.png', dpi=200, bbox_inches='tight')

def compare_feature_computation_time(use_angular = True, calculate_gradient = False):
    print('compare_feature_computation_time')

    from tqdm import trange
    from ase.spacegroup import crystal

    a = 4.6
    c = 2.95
    rutile = crystal(['Ti', 'O'], basis=[(0, 0, 0), (0.3, 0.3, 0.0)],
                    spacegroup=136, cellpar=[a, a, c, 90, 90, 90])

    mol = molecule('CH3CH2OH')
    mol.center(5)

    systems = [mol, rutile]
    
    
    eta = 5
    r_cut_all = 5.0
    n_bins = 30

    n_runs = 1000

    for atoms in systems:
        print(atoms.get_pbc(), len(atoms), atoms.get_chemical_formula())
        trajs = []
        descriptor = JaxFingerprint.from_atoms(atoms, r_cut_radial=r_cut_all, r_cut_angular=r_cut_all,
                                         n_bins_radial=n_bins, use_angular=use_angular, eta=eta)

        from agox.models.descriptors.fingerprint_cython.angular_fingerprintFeature_cy import Angular_Fingerprint
        descriptor_cython = Angular_Fingerprint(atoms, Rc1=r_cut_all, Rc2=r_cut_all, binwidth1=r_cut_all/n_bins, nsigma=10,
                                            sigma1=0.2, sigma2=0.2, Nbins2=n_bins,
                                            eta=eta, gamma=2, use_angular=use_angular)

        for i in range(n_runs):
            new_atoms = atoms.copy()
            new_atoms.rattle(0.01, seed=i)
            trajs.append(new_atoms)
        
        print('jax-Fingerprint')
        for i in trange(n_runs):
            atoms.set_positions(trajs[i].get_positions())
            features = descriptor.create_features(atoms)
        
        print('cython-Fingerprint')
        for i in trange(n_runs):
            atoms.set_positions(trajs[i].get_positions())
            features = descriptor_cython.get_feature(atoms)
            if calculate_gradient:
                feature_grad = descriptor_cython.get_featureGradient(atoms)

if __name__ == '__main__':
    main()
    compare_feature_computation_time()



import numpy as np
from scipy.linalg import svd

from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass

class CUR(SparsifierBaseClass):
    name = "CUR"

    def sparsify(self, X: np.ndarray) -> np.ndarray:
        if X.shape[0] < self.m_points:
            m_indices = np.arange(0, X.shape[0])
            return X, m_indices

        U, _, _ = svd(X)

        fd = np.min(X.shape)
        score = np.sum(U[:, :fd] ** 2, axis=1) / fd

        # Pick probabilisticly according to the score
        indices = np.random.choice(np.arange(0, X.shape[0]), self.m_points, p=score, replace=False)
        Xm = X[indices, :]

        return Xm, indices



import numpy as np

from agox.generators.block_generators.ABC_block import BlockGeneratorBaseClass


class RandomBlockGenerator(BlockGeneratorBaseClass):
    name = "RandomBlockGenerator"

    def _get_candidates(self, candidate, parents, environment):
        template = environment.get_template()

        blocks_placed = np.zeros(len(self.building_blocks))
        block_indices = []
        blocks_used = []
        while np.sum(blocks_placed) < np.sum(self.N_blocks):
            block, block_index = self.get_building_block(
                blocks_placed, random_rotation=True
            )

            if (blocks_placed == 0).all():
                suggested_position = self.get_box_vector()
            else:
                r_min = 1
                r_max = 3.5
                radius = np.random.rand() * (r_max - r_min) + r_min
                displacement = self.get_displacement_vector(radius)
                xyz0 = candidate.positions[
                    np.random.randint(low=0, high=len(candidate), size=1)
                ].reshape(3)
                suggested_position = xyz0 + displacement

                if not self.check_confinement(suggested_position):
                    continue

            block.positions += suggested_position

            # Checks confinement for all atoms in the building block at once.
            if not self.check_confinement(block.positions).any():
                continue

            # Checks that distances are not too bad.
            # This implementation does NOT care about PBCs.
            if not self.check_distances(candidate, block, present=False):
                continue

            block_indices += [(np.arange(len(candidate), len(candidate) + len(block)))]
            candidate += block
            blocks_placed[block_index] += 1
            blocks_used.append(block_index)

        candidate.add_meta_information("blocks_used", np.array(blocks_used))
        candidate.add_meta_information(
            "block_indices", np.array(block_indices, dtype=object)
        )

        if self.apply_constraint:
            constraint = self.get_fix_internal_constraint(blocks_used)
            candidate.set_constraint(constraint)

        return [candidate]

    def get_number_of_parents(self, sampler):
        return 0



from typing import List
from warnings import warn

import numpy as np
from ase import Atoms

from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass
from agox.utils.filters.ABC_filter import FilterBaseClass
from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass


class SparsifierFilter(FilterBaseClass):

    """
    Filter based on a sparsifier.

    Uses a descriptor and a sparsifier to filter the data by selecting 
    features. The descriptor is used to compute the features and the
    sparsifier is used to select the features.

    Parameters
    ----------
    sparsifier : SparsifierBaseClass
        Sparsifier to use
    descriptor : DescriptorBaseClass, optional
        Descriptor to use.        
    """

    def __init__(
        self,
        sparsifier: SparsifierBaseClass,
        descriptor: DescriptorBaseClass = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sparsifier = sparsifier
        if descriptor is not None:
            self.descriptor = descriptor
            self.feature_method = self.descriptor.get_features
        else:
            warn("Using indexes as features")
            self.feature_method = lambda atoms: np.arange(len(atoms))

    def _filter(self, atoms: List[Atoms]) -> np.ndarray:
        X = self.preprocess(atoms)
        _, idxs = self.sparsifier(X)
        return idxs

    def preprocess(self, atoms: List[Atoms] = None) -> np.ndarray:
        """
        Preprocess the data by computing the features.

        Parameters
        ----------
        atoms : List[Atoms]
            List of atoms objects

        Returns
        -------
        X : np.ndarray
            Matrix with rows corresponding to features

        """

        f = self.feature_method(atoms)
        if isinstance(f, np.ndarray) and len(f.shape) == 1:
            f = f.reshape(1, -1)
        f = np.vstack(f)
        return f

    @property
    def name(self):
        return f"{self.sparsifier.name}-Filter"
